# encoding: utf-8
# This file is part of Checkbox.
#
# Copyright 2019 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.
"""
Definition for UnifiedRunner class.
"""

import contextlib
import getpass
import gzip
import io
import logging
import os
import select
import subprocess
import sys
import tempfile
import threading
import time

from plainbox.abc import IJobResult, IJobRunner
from plainbox.i18n import gettext as _
from plainbox.impl.color import Colorizer
from plainbox.impl.unit.job import supported_plugins
from plainbox.impl.unit.unit import on_ubuntucore
from plainbox.impl.result import IOLogRecordWriter
from plainbox.impl.result import JobResultBuilder
from plainbox.impl.runner import CommandOutputWriter
from plainbox.impl.runner import IOLogRecordGenerator
from plainbox.impl.runner import JobRunnerUIDelegate
from plainbox.impl.runner import slugify
from plainbox.impl.jobcache import ResourceJobCache
from plainbox.impl.secure.sudo_broker import sudo_password_provider
from plainbox.impl.session.storage import WellKnownDirsHelper
from plainbox.vendor import extcmd

logger = logging.getLogger("plainbox.unified")


class UnifiedRunner(IJobRunner):
    """
    Class for handling running of jobs.

    Instance of this class should is responsible for creating proper
    environment for job's command can run in.
    """

    def __init__(self, session_id, provider_list, jobs_io_log_dir,
                 command_io_delegate=None, dry_run=False,
                 execution_ctrl_list=None, stdin=False,
                 normal_user_provider=lambda: None,
                 password_provider=sudo_password_provider.get_sudo_password,
                 extra_env=None):
        self._session_id = session_id
        self._provider_list = provider_list
        if execution_ctrl_list is not None:
            logger.info("Using custom execution controllers is deprecated")
        self._jobs_io_log_dir = jobs_io_log_dir
        self._job_runner_ui_delegate = JobRunnerUIDelegate()
        self._command_io_delegate = command_io_delegate
        self._dry_run = dry_run
        self._resource_cache = ResourceJobCache()
        self._resource_cache.load()
        self._user_provider = normal_user_provider
        self._password_provider = password_provider
        self._stdin = stdin
        self._running_jobs_pid = None
        self._extra_env = extra_env

    def run_job(self, job, job_state, environ=None, ui=None):
        logger.info(_("Running %r"), job)
        if job.plugin not in supported_plugins:
            print(Colorizer().RED("Unsupported plugin type: {}".format(
                job.plugin)))
            return JobResultBuilder(
                outcome=IJobResult.OUTCOME_SKIP,
                comments=_("Unsupported plugin type: {}".format(job.plugin))
            ).get_result()

        # resource and attachment jobs are always run (even in dry runs)
        if self._dry_run and job.plugin not in ('resource', 'attachment'):
            return JobResultBuilder(
                outcome=IJobResult.OUTCOME_SKIP,
                comments=_("Job skipped in dry-run mode")
            ).get_result()
        self._job_runner_ui_delegate.ui = ui

        # for cached resource jobs we get the result using cache
        # if it's not in the cache, ordinary "_run_command" will be run
        if job.plugin == 'resource' and 'cachable' in job.get_flag_set():
            from_cache, result = self._resource_cache.get(
                job.checksum, lambda: self._run_command(
                    job, environ).get_result())
            if from_cache:
                print(Colorizer().header(_("Using cached data!")))
                jrud = self._job_runner_ui_delegate
                jrud.on_begin('', dict())
                for io_log_entry in result.io_log:
                    jrud.on_chunk(io_log_entry.stream_name, io_log_entry.data)
                jrud.on_end(result.return_code)
            return result

        # manual jobs don't require running anything so we just return
        # the 'undecided' outcome
        if job.plugin == 'manual':
            return JobResultBuilder(
                outcome=IJobResult.OUTCOME_UNDECIDED).get_result()

        # all other kinds of jobs at this point need to run their command
        if not job.command:
            print(Colorizer().RED("No command to run!"))
            return JobResultBuilder(
                outcome=IJobResult.OUTCOME_FAIL,
                comments=_("No command to run!")
            ).get_result()
        result_builder = self._run_command(job, environ)

        # for user-interact-verify and user-verify jobs the operator chooses
        # the final outcome, so we need to reset the outcome to undecided
        # (from what command's return code would have set)
        if job.plugin in ('user-interact-verify', 'user-verify'):
            result_builder.outcome = IJobResult.OUTCOME_UNDECIDED

        # by this point the result_builder should have all the info needed
        # to yield appropriate result
        return result_builder.get_result()

    def get_warm_up_sequence(self, job_list):
        # we no longer need a warm-up sequence
        # this is left here to conform to the interface
        return []

    def _run_command(self, job, environ):
        start_time = time.time()
        slug = slugify(job.id)
        output_writer = CommandOutputWriter(
            stdout_path=os.path.join(
                self._jobs_io_log_dir, "{}.stdout".format(slug)),
            stderr_path=os.path.join(
                self._jobs_io_log_dir, "{}.stderr".format(slug)))
        io_log_gen = IOLogRecordGenerator()
        log = os.path.join(self._jobs_io_log_dir, "{}.record.gz".format(slug))
        with gzip.open(log, mode='wb') as gzip_stream, io.TextIOWrapper(
                gzip_stream, encoding='UTF-8') as record_stream:
            writer = IOLogRecordWriter(record_stream)
            io_log_gen.on_new_record.connect(writer.write_record)
            delegate = extcmd.Chain([
                self._job_runner_ui_delegate, io_log_gen,
                self._command_io_delegate, output_writer])
            ecmd = extcmd.ExternalCommandWithDelegate(delegate)
            return_code = self.execute_job(job, environ, ecmd, self._stdin)
            io_log_gen.on_new_record.disconnect(writer.write_record)
        if return_code == 0:
            outcome = IJobResult.OUTCOME_PASS
        elif return_code < 0:
            outcome = IJobResult.OUTCOME_CRASH
        else:
            outcome = IJobResult.OUTCOME_FAIL
        return JobResultBuilder(
            outcome=outcome,
            return_code=return_code,
            io_log_filename=log,
            execution_duration=time.time() - start_time)

    def execute_job(self, job, environ, extcmd_popen, stdin=None):
        """Run the 'binary' associated with the job."""
        target_user = job.user or self._user_provider()
        if target_user == getpass.getuser():
            target_user = None

        def call(extcmd_popen, *args, **kwargs):
            """Handle low-level subprocess stuff."""
            is_alive = True
            # Notify that the process is about to start
            extcmd_popen._delegate.on_begin(args, kwargs)
            # Setup stdout/stderr redirection
            kwargs['stdout'] = subprocess.PIPE
            kwargs['stderr'] = subprocess.PIPE
            kwargs['start_new_session'] = True
            # Prepare stdio supply
            in_r, in_w = os.pipe()
            # first let's punch the password in
            # we need it only if the target user differs from the one that
            # started checkbox and when changing the user (sudo) requires
            # password
            if target_user and self._password_provider:
                password = self._password_provider()
                if password:
                    os.write(in_w, password + b'\n')

            def stdin_forwarder(stdin):
                """Forward data from one pipe to the other."""
                # use systems stdin if the stdin pipe wasn't provided
                stdin = stdin or sys.stdin
                try:
                    while is_alive:
                        if stdin in select.select([stdin], [], [], 0)[0]:
                            buf = stdin.readline()
                            if buf == '':
                                break
                            os.write(in_w, buf.encode(stdin.encoding))
                        else:
                            time.sleep(0.1)
                except BrokenPipeError:
                    pass
                os.close(in_w)
            forwarder_thread = threading.Thread(
                target=stdin_forwarder, args=(stdin,))
            forwarder_thread.start()
            kwargs['stdin'] = in_r

            # Start the process
            proc = extcmd_popen._popen(*args, **kwargs)
            self._running_jobs_pid = proc.pid
            # Setup all worker threads. By now the pipes have been created and
            # proc.stdout/proc.stderr point to open pipe objects.
            stdout_reader = threading.Thread(
                target=extcmd_popen._read_stream, args=(proc.stdout, "stdout"))
            stderr_reader = threading.Thread(
                target=extcmd_popen._read_stream, args=(proc.stderr, "stderr"))
            queue_worker = threading.Thread(target=extcmd_popen._drain_queue)
            # Start all workers
            queue_worker.start()
            stdout_reader.start()
            stderr_reader.start()
            try:
                while True:
                    try:
                        proc.wait()
                        break
                    except KeyboardInterrupt:
                        is_alive = False
                        import signal
                        self.send_signal(signal.SIGKILL, target_user)
                        # And send a notification about this
                        extcmd_popen._delegate.on_interrupt()
            finally:
                self._running_jobs_pid = None
                # Wait until all worker threads shut down
                stdout_reader.join()
                proc.stdout.close()
                stderr_reader.join()
                proc.stderr.close()
                # Tell the queue worker to shut down
                extcmd_popen._queue.put(None)
                queue_worker.join()
                os.close(in_r)
                is_alive = False
                forwarder_thread.join()
            # Notify that the process has finished
            extcmd_popen._delegate.on_end(proc.returncode)
            return proc.returncode
        # Setup the executable nest directory
        with self.configured_filesystem(job) as nest_dir:
            # Get the command and the environment.
            # of this execution controller
            cmd = get_execution_command(job, environ, self._session_id,
                                        nest_dir, target_user, self._extra_env)
            env = get_execution_environment(job, environ, self._session_id,
                                            nest_dir)
            if self._user_provider():
                env['NORMAL_USER'] = self._user_provider()
            # run the command
            logger.debug(_("job[%(ID)s] executing %(CMD)r with env %(ENV)r"),
                         {"ID": job.id, "CMD": cmd,
                          "ENV": env})
            if 'preserve-cwd' in job.get_flag_set() or os.getenv("SNAP"):
                return_code = call(
                    extcmd_popen, cmd, stdin=subprocess.PIPE, env=env)
            else:
                with self.temporary_cwd(job) as cwd_dir:
                    return_code = call(
                        extcmd_popen, cmd, stdin=subprocess.PIPE, env=env,
                        cwd=cwd_dir)
            if 'noreturn' in job.get_flag_set():
                import signal
                signal.pause()
            return return_code

    @contextlib.contextmanager
    def configured_filesystem(self, job):
        """
        Context manager for handling filesystem aspects of job execution.

        :param job:
            The JobDefinition to execute
        :returns:
            Pathname of the executable symlink nest directory.
        """
        # Create a nest for all the private executables needed for execution
        prefix = 'nest-'
        suffix = '.{}'.format(job.checksum)
        with tempfile.TemporaryDirectory(suffix, prefix) as nest_dir:
            os.chmod(nest_dir, 0o777)
            logger.debug(_("Symlink nest for executables: %s"), nest_dir)
            from plainbox.impl.ctrl import SymLinkNest
            nest = SymLinkNest(nest_dir)
            # Add all providers sharing namespace with the current job to PATH
            for provider in self._provider_list:
                if job.provider.namespace == provider.namespace:
                    nest.add_provider(provider)
            yield nest_dir

    @contextlib.contextmanager
    def temporary_cwd(self, job):
        """
        Context manager for handling temporary current working directory
        for a particular execution of a job definition command.

        :param job:
            The JobDefinition to execute
        :returns:
            Pathname of the new temporary directory
        """
        # Create a nest for all the private executables needed for execution
        prefix = 'cwd-'
        suffix = '.{}'.format(job.checksum)
        try:
            with tempfile.TemporaryDirectory(suffix, prefix) as cwd_dir:
                logger.debug(
                    _("Job temporary current working directory: %s"), cwd_dir)
                try:
                    yield cwd_dir
                finally:
                    if 'has-leftovers' not in job.get_flag_set():
                        self._check_leftovers(cwd_dir, job)

        except PermissionError as exc:
            logger.warning(
                _("There was a problem with temporary cwd %s, %s"),
                cwd_dir, exc)

    def _check_leftovers(self, cwd_dir, job):
        leftovers = []
        for dirpath, dirnames, filenames in os.walk(cwd_dir):
            if dirpath != cwd_dir:
                leftovers.append(dirpath)
            leftovers.extend(
                os.path.join(dirpath, filename)
                for filename in filenames)
        if leftovers:
            logger.warning(
                _("Job {0} created leftover filesystem artefacts"
                  " in its working directory").format(job.id))
            for item in leftovers:
                logger.warning(_(
                    "Leftover file/directory: %r"),
                    os.path.relpath(item, cwd_dir))
            logger.warning(
                _("Please store desired files in $PLAINBOX_SESSION_SHARE"
                    "and use regular temporary files for everything else"))

    def get_record_path_for_job(self, job):
        return os.path.join(self._jobs_io_log_dir,
                            "{}.record.gz".format(slugify(job.id)))

    def send_signal(self, signal, target_user):
        if not target_user:
            os.kill(self._running_jobs_pid, signal)
        else:
            # process used sudo, so sudo is needed to kill it
            in_r, in_w = os.pipe()
            os.write(in_w, self._password_provider() + b'\n')
            cmd = ['sudo', '--prompt', '', '--reset-timestamp', '--stdin',
                   '--user', 'root', 'kill', '-s', str(signal),
                   '-{}'.format(self._running_jobs_pid)]
            try:
                subprocess.check_call(cmd, stdin=in_r)
            except subprocess.CalledProcessError:
                logger.warning("Failed to kill process")


class FakeJobRunner(UnifiedRunner):
    """
    Fake runner for jobs.

    Special runner that creates fake resource objects.
    """

    def run_job(self, job, job_state, environ=None, ui=None):
        """
        Only one resouce object is created from this runner.
        Exception: 'graphics_card' resource job creates two objects to
        simulate hybrid graphics.
        """
        if job.plugin != 'resource':
            return super().run_job(job, job_state, environ, ui)
        builder = JobResultBuilder()
        if job.partial_id == 'graphics_card':
            builder.io_log = [(0, 'stdout', b'a: b\n'),
                              (1, 'stdout', b'\n'),
                              (2, 'stdout', b'a: c\n')]
        else:
            builder.io_log = [(0, 'stdout', b'a: b\n')]
        builder.outcome = 'pass'
        builder.return_code = 0
        return builder.get_result()


def get_execution_environment(job, environ, session_id, nest_dir):
    """
    Get the environment required to execute the specified job:

    :param job:
        job definition with the command and environment definitions
    :param environ:
        A dictionary of environment variables which can be used to load missing
        environment definitions that apply to all jobs. It is used to provide
        values for missing environment variables that are required by the job
        (as expressed by the environ key in the job definition file).
    :param session_id:
        ID of the session that will be used to retrieve the location of session
        data
    :param nest_dir:
        A directory with a nest of symlinks to all executables required to
        execute the specified job. This argument may or may not be used,
        depending on how PATH is passed to the command (via environment or via
        the commant line)
    :return:
        dictionary with the environment to use.
    """
    # Get a proper environment
    env = dict(os.environ)
    if 'reset-locale' in job.get_flag_set():
        # Use non-internationalized environment
        env['LANG'] = 'C.UTF-8'
        if 'LANGUAGE' in env:
            del env['LANGUAGE']
        for name in list(env.keys()):
            if name.startswith("LC_"):
                del env[name]
    else:
        # Set the per-provider gettext domain and locale directory
        if job.provider.gettext_domain is not None:
            env['TEXTDOMAIN'] = env['PLAINBOX_PROVIDER_GETTEXT_DOMAIN'] = \
                job.provider.gettext_domain
        if job.provider.locale_dir is not None:
            env['TEXTDOMAINDIR'] = env['PLAINBOX_PROVIDER_LOCALE_DIR'] = \
                job.provider.locale_dir
        if (os.getenv("SNAP") or os.getenv("SNAP_APP_PATH")):
            copy_vars = ['PYTHONHOME', 'PYTHONUSERBASE', 'LD_LIBRARY_PATH',
                         'GI_TYPELIB_PATH', 'PERL5LIB']
            for key, value in env.items():
                if key in copy_vars or key.startswith('SNAP'):
                    env[key] = value
    # Use PATH that can lookup checkbox scripts
    if job.provider.extra_PYTHONPATH:
        env['PYTHONPATH'] = os.pathsep.join(
            [job.provider.extra_PYTHONPATH] + env.get(
                "PYTHONPATH", "").split(os.pathsep))
    # Inject nest_dir into PATH
    env['PATH'] = os.pathsep.join(
        [nest_dir] + env.get("PATH", "").split(os.pathsep))
    # Add per-session shared state directory
    env['PLAINBOX_SESSION_SHARE'] = WellKnownDirsHelper.session_share(
        session_id)

    def set_if_not_none(envvar, source):
        """Update env if the source variable is not None"""
        if source is not None:
            env[envvar] = source
    set_if_not_none('PLAINBOX_PROVIDER_DATA', job.provider.data_dir)
    set_if_not_none('PLAINBOX_PROVIDER_UNITS', job.provider.units_dir)
    set_if_not_none('CHECKBOX_SHARE', job.provider.CHECKBOX_SHARE)
    # Inject additional variables that are requested in the config
    if environ is not None:
        for env_var in environ:
            # Don't override anything that is already present in the
            # current environment. This will allow users to customize
            # variables without editing any config files.
            if env_var in env:
                continue
            # The environ dict is populated from the environment section of the
            # config file. If it has a particular variable then copy it over.
            env[env_var] = environ[env_var]
    return env


def get_differential_execution_environment(job, environ, session_id, nest_dir,
                                           extra_env=None):
    """
    Get the environment required to execute the specified job:

    :param job:
        job definition with the command and environment definitions
    :param environ:
        A dictionary of environment variables which can be used to load missing
        environment definitions that apply to all jobs. It is used to
        provide values for missing environment variables that are required
        by the job (as expressed by the environ key in the job definition
        file).
    :param session_id:
        ID of the session that will be used to retrieve the location of session
        data
    :param nest_dir:
        A directory with a nest of symlinks to all executables required to
        execute the specified job. This is simply passed to
        :meth:`get_execution_environment()` directly.
    :returns:
        Differential environment (see below).

    This implementation computes the desired environment (as it was
    computed in the base class) and then discards all of the environment
    variables that are identical in both sets. The exception are variables
    that are mentioned in
    :meth:`plainbox.impl.job.JobDefinition.get_environ_settings()` which
    are always retained.
    """
    base_env = os.environ
    target_env = get_execution_environment(job, environ, session_id, nest_dir)
    delta_env = {
        key: value
        for key, value in target_env.items()
        if key not in base_env or base_env[key] != value
        or key in job.get_environ_settings()
    }
    if 'reset-locale' in job.get_flag_set():
        delta_env['LANG'] = 'C.UTF-8'
        delta_env['LANGUAGE'] = ''
        delta_env['LC_ALL'] = 'C.UTF-8'
    if extra_env:
        delta_env.update(extra_env())
    # Preserve the copy_vars variables + those prefixed with SNAP on Snappy
    if (os.getenv("SNAP") or os.getenv("SNAP_APP_PATH")):
        copy_vars = ['PYTHONHOME', 'PYTHONUSERBASE', 'LD_LIBRARY_PATH',
                     'GI_TYPELIB_PATH', 'PROVIDERPATH', 'PYTHONPATH',
                     'PERL5LIB']
        for key, value in base_env.items():
            if key in copy_vars or key.startswith('SNAP'):
                delta_env[key] = value
    return delta_env


def get_execution_command(job, environ, session_id,
                          nest_dir, target_user=None, extra_env=None):
    """Generate a command argv to run in the shell."""
    cmd = []
    if target_user:
        # we want sudo to:
        #   - have no prompt (--prompt '')
        #   - reset the timestamp, so it predictably asks for password
        #     (--reset-timestamp)
        #   - gets password as the first line of stdin (--stdin)
        #   - change the user to the target user (--user)
        cmd = ['sudo', '--prompt', '', '--reset-timestamp', '--stdin',
               '--user', target_user]
    cmd += ['env']
    if target_user:
        env = get_differential_execution_environment(
            job, environ, session_id, nest_dir, extra_env)
    else:
        env = get_execution_environment(job, environ, session_id, nest_dir)
        if extra_env:
            env.update(extra_env())
    cmd += ["{key}={value}".format(key=key, value=value)
            for key, value in sorted(env.items())]
    # Run the command unconfined on ubuntu core because of snap-confine fixes
    # related to https://ubuntu.com/security/CVE-2021-44731
    if on_ubuntucore():
        cmd += ['aa-exec', '-p', 'unconfined']
    cmd += [job.shell, '-c', job.command]
    return cmd



# This file is part of Checkbox.
#
# Copyright 2012, 2013 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`plainbox.impl.session.test_storage`
=========================================

Test definitions for :mod:`plainbox.impl.session.storage`
"""

from unittest import TestCase
import os

from plainbox.impl.session.storage import SessionStorage
from plainbox.impl.session.storage import WellKnownDirsHelper


class SessionStorageTests(TestCase):

    def test_smoke(self):
        session_prefix = "test_storage-"
        storage = SessionStorage(session_prefix)
        session_id = storage.id
        self.assertEqual(storage.location,
                         WellKnownDirsHelper.session_dir(session_id))

    def test_create_remove(self):
        session_prefix = "test_storage-"
        # Create a new storage in the specified directory
        storage = SessionStorage.create(session_prefix)
        session_id = storage.id
        # The location should have been created
        self.assertTrue(os.path.exists(storage.location))
        # And it should be in the directory we indicated
        self.assertEqual(os.path.dirname(storage.location),
                         os.path.dirname(WellKnownDirsHelper.session_dir(
                             session_id)))
        # Remove the storage now
        storage.remove()
        # And make sure the storage is gone
        self.assertFalse(os.path.exists(storage.location))

    def test_load_save_checkpoint(self):
        session_prefix = "test_storage-"
        # Create a new storage in the specified directory
        storage = SessionStorage.create(session_prefix)
        # Save some checkpoint data
        data_out = b'some data'
        storage.save_checkpoint(data_out)
        # Load it back
        data_in = storage.load_checkpoint()
        # Check if it's right
        self.assertEqual(data_out, data_in)
        # Remove the storage now
        storage.remove()



# This file is part of Checkbox.
#
# Copyright 2018 Canonical Ltd.
# Authors:
#     Sylvain Pineau <sylvain.pineau@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`checkbox-ng.launcher.merge_submissions` -- merge-submissions sub-command
==============================================================================
"""
import tarfile
from tempfile import TemporaryDirectory

from checkbox_ng.launcher.merge_reports import MergeReports
from plainbox.impl.session import SessionManager


class MergeSubmissions(MergeReports):
    name = 'merge-submissions'

    def register_arguments(self, parser):
        parser.add_argument(
            'submission', nargs='*', metavar='SUBMISSION',
            help='submission tarball')
        parser.add_argument(
            '-o', '--output-file', metavar='FILE',  required=True,
            help='save combined test results to the specified FILE')
        parser.add_argument(
            '--title', action='store', metavar='SESSION_NAME',
            help='title of the session to use')

    def invoked(self, ctx):
        tmpdir = TemporaryDirectory()
        self.job_dict = {}
        self.category_dict = {}
        for submission in ctx.args.submission:
            session_title = self._parse_submission(
                submission, tmpdir, mode='dict')
        manager = SessionManager.create_with_unit_list(
            list(self.job_dict.values()) + list(self.category_dict.values()))
        manager.state.metadata.title = ctx.args.title or session_title
        for job in self.job_dict.values():
            self._populate_session_state(job, manager.state)
        exporter = self._create_exporter(
            'com.canonical.plainbox::tar')
        with open(ctx.args.output_file, 'wb') as stream:
            exporter.dump_from_session_manager(manager, stream)
        with tarfile.open(ctx.args.output_file) as tar:
            tar.extractall(tmpdir.name)
        with tarfile.open(ctx.args.output_file, mode='w:xz') as tar:
            tar.add(tmpdir.name, arcname='')
        print(ctx.args.output_file)



# This file is part of Checkbox.
#
# Copyright 2017-2019 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.
"""
This module contains implementation of the master end of the remote execution
functionality.
"""
import contextlib
import getpass
import gettext
import ipaddress
import json
import logging
import os
import select
import socket
import time
import signal
import sys

from collections import namedtuple
from functools import partial
from tempfile import SpooledTemporaryFile

from plainbox.impl.color import Colorizer
from plainbox.impl.launcher import DefaultLauncherDefinition
from plainbox.impl.secure.config import Unset
from plainbox.impl.session.remote_assistant import RemoteSessionAssistant
from plainbox.vendor import rpyc
from checkbox_ng.urwid_ui import TestPlanBrowser
from checkbox_ng.urwid_ui import CategoryBrowser
from checkbox_ng.urwid_ui import ManifestBrowser
from checkbox_ng.urwid_ui import ReRunBrowser
from checkbox_ng.urwid_ui import interrupt_dialog
from checkbox_ng.urwid_ui import resume_dialog
from checkbox_ng.launcher.run import NormalUI, ReRunJob
from checkbox_ng.launcher.stages import MainLoopStage
from checkbox_ng.launcher.stages import ReportsStage
from tqdm import tqdm
_ = gettext.gettext
_logger = logging.getLogger("master")


class SimpleUI(NormalUI, MainLoopStage):
    """
    Simplified version of the NormalUI from checkbox_ng.launcher.run.

    The simplification is mainly about just dealing with text that is to be
    displayed, instead of the plainbox abstractions like job, job state, etc.

    It's a class just for namespacing purposes.
    """

    C = Colorizer()

    # XXX: evaluate other ways of aggregating those functions

    def description(header, text):
        print(SimpleUI.C.WHITE(header))
        print()
        print(SimpleUI.C.CYAN(text))
        print()

    def header(header):
        print(SimpleUI.C.header(header, fill='-'))

    def green_text(text, end='\n'):
        print(SimpleUI.C.GREEN(text), end=end, file=sys.stdout)

    def red_text(text, end='\n'):
        print(SimpleUI.C.RED(text), end=end, file=sys.stderr)

    def black_text(text, end='\n'):
        print(SimpleUI.C.BLACK(text), end=end, file=sys.stdout)

    def horiz_line():
        print(SimpleUI.C.WHITE('-' * 80))

    @property
    def is_interactive(self):
        return True

    @property
    def sa(self):
        None


class RemoteMaster(ReportsStage, MainLoopStage):
    """
    Control remote slave instance

    This class implements the part that presents UI to the operator and
    steers the session.
    """

    name = 'remote-control'

    @property
    def is_interactive(self):
        return (self.launcher.ui_type == 'interactive' and
                sys.stdin.isatty() and sys.stdout.isatty())

    @property
    def C(self):
        return self._C

    @property
    def sa(self):
        return self._sa

    def invoked(self, ctx):
        self._C = Colorizer()
        self._override_exporting(self.local_export)
        self._launcher_text = ''
        self._is_bootstrapping = False
        self._target_host = ctx.args.host
        self._normal_user = ''
        self.launcher = DefaultLauncherDefinition()
        if ctx.args.launcher:
            expanded_path = os.path.expanduser(ctx.args.launcher)
            if not os.path.exists(expanded_path):
                raise SystemExit(_("{} launcher file was not found!").format(
                    expanded_path))
            with open(expanded_path, 'rt') as f:
                self._launcher_text = f.read()
            self.launcher.read_string(self._launcher_text)
        if ctx.args.user:
            self._normal_user = ctx.args.user
        timeout = 600
        deadline = time.time() + timeout
        port = ctx.args.port
        if not ipaddress.ip_address(ctx.args.host).is_loopback:
            print(_("Connecting to {}:{}. Timeout: {}s").format(
                ctx.args.host, port, timeout))
        while time.time() < deadline:
            try:
                self.connect_and_run(ctx.args.host, port)
                break
            except (ConnectionRefusedError, socket.timeout, OSError):
                print('.', end='', flush=True)
                time.sleep(1)
        else:
            print(_("\nConnection timed out."))

    def connect_and_run(self, host, port=18871):
        config = rpyc.core.protocol.DEFAULT_CONFIG.copy()
        config['allow_all_attrs'] = True
        config['sync_request_timeout'] = 120
        keep_running = False
        server_msg = None
        self._prepare_transports()
        interrupted = False
        while True:
            try:
                if interrupted:
                    _logger.info("remote: Session interrupted")
                    interrupted = False  # we are handling the interruption ATM
                    # next line can raise exception due to connection being
                    # lost so let's set the default behavior to quitting
                    keep_running = False
                    keep_running = self._handle_interrupt()
                    if not keep_running:
                        break
                conn = rpyc.connect(host, port, config=config)
                keep_running = True

                def quitter(msg):
                    # this will be called when the slave decides to disconnect
                    # this master
                    nonlocal server_msg
                    nonlocal keep_running
                    keep_running = False
                    server_msg = msg
                with contextlib.suppress(AttributeError):
                    # TODO: REMOTE_API
                    # when bumping the remote api make this bit obligatory
                    # i.e. remove the suppressing
                    conn.root.register_master_blaster(quitter)
                self._sa = conn.root.get_sa()
                self.sa.conn = conn
                # TODO: REMOTE API RAPI: Remove this API on the next RAPI bump
                # the check and bailout is not needed if the slave as up to
                # date as this master, so after bumping RAPI we can assume
                # that slave is always passwordless
                if not self.sa.passwordless_sudo:
                    raise SystemExit(
                        _("This version of Checkbox requires the service"
                          " to be run as root"))
                try:
                    slave_api_version = self.sa.get_remote_api_version()
                except AttributeError:
                    raise SystemExit(_("Service doesn't declare Remote API"
                                       " version. Update Checkbox on the"
                                       " SUT!"))
                master_api_version = RemoteSessionAssistant.REMOTE_API_VERSION
                if slave_api_version != master_api_version:
                    raise SystemExit(_("Remote API version mismatch. Service "
                                       "uses: {}. Remote uses: {}").format(
                        slave_api_version, master_api_version))
                state, payload = self.sa.whats_up()
                _logger.info("remote: Main dispatch with state: %s", state)
                keep_running = {
                    'idle': self.new_session,
                    'running': self.wait_and_continue,
                    'finalizing': self.finish_session,
                    'testsselected': partial(
                        self.run_jobs, resumed_session_info=payload),
                    'bootstrapping': self.restart,
                    'bootstrapped': partial(
                        self.select_jobs, all_jobs=payload),
                    'started': self.restart,
                    'interacting': partial(
                        self.resume_interacting, interaction=payload),
                }[state]()
            except EOFError as exc:
                if keep_running:
                    print("Connection lost!")
                    # this is yucky but it works, in case of explicit
                    # connection closing by the slave we get this msg
                    _logger.info("remote: Connection lost due to: %s", exc)
                    if str(exc) == 'stream has been closed':
                        print('Service explicitly disconnected you. Possible '
                              'reason: new remote connected to the service')
                        break
                    print(exc)
                    time.sleep(1)
                else:
                    # if keep_running got set to False it means that the
                    # network interruption was planned, AKA slave disconnected
                    # this master
                    print(server_msg)
                    break
            except (ConnectionRefusedError, socket.timeout, OSError) as exc:
                _logger.info("remote: Connection lost due to: %s", exc)
                if not keep_running:
                    raise
                # it's reconnecting, so we can ignore refuses
                print('Reconnecting...')
                time.sleep(0.5)
            except KeyboardInterrupt:
                interrupted = True

            if not keep_running:
                break

    def new_session(self):
        _logger.info("remote: Starting new session.")
        configuration = dict()
        configuration['launcher'] = self._launcher_text
        configuration['normal_user'] = self._normal_user

        try:
            tps = self.sa.start_session(configuration)
        except RuntimeError as exc:
            raise SystemExit(exc.args[0]) from exc
        if self.launcher.test_plan_forced:
            self.select_tp(self.launcher.test_plan_default_selection)
            self.select_jobs(self.jobs)
        else:
            self.interactively_choose_tp(tps)

    def interactively_choose_tp(self, tps):
        _logger.info("remote: Interactively choosing TP.")
        tp_info_list = [{'id': tp[0], 'name': tp[1]} for tp in tps]
        if not tp_info_list:
            print(_("There were no test plans to select from!"))
            raise SystemExit(0)
        selected_tp = TestPlanBrowser(
            _("Select test plan"),
            tp_info_list, None).run()
        if selected_tp is None:
            print(_("Nothing selected"))
            raise SystemExit(0)

        self.select_tp(selected_tp)
        if not self.jobs:
            print(self.C.RED(_("There were no tests to select from!")))
            self.sa.finalize_session()
            return
        self.select_jobs(self.jobs)

    def select_tp(self, tp):
        _logger.info("remote: Selected test plan: %s", tp)
        self.sa.prepare_bootstrapping(tp)
        self._is_bootstrapping = True
        bs_todo = self.sa.get_bootstrapping_todo_list()
        for job_no, job_id in enumerate(bs_todo, start=1):
            print(self.C.header(
                _('Bootstrap {} ({}/{})').format(
                    job_id, job_no, len(bs_todo), fill='-')))
            self.sa.run_bootstrapping_job(job_id)
            self.wait_for_job()
        self._is_bootstrapping = False
        self.jobs = self.sa.finish_bootstrap()

    def _strtobool(self, val):
        return val.lower() in ('y', 'yes', 't', 'true', 'on', '1')

    def select_jobs(self, all_jobs):
        if self.launcher.test_selection_forced:
            if self.launcher.manifest is not Unset:
                self.sa.save_manifest(
                    {manifest_id:
                     self._strtobool(self.launcher.manifest[manifest_id]) for
                     manifest_id in self.launcher.manifest}
                )
        else:
            _logger.info("remote: Selecting jobs.")
            reprs = json.loads(self.sa.get_jobs_repr(all_jobs))
            wanted_set = CategoryBrowser(
                "Choose tests to run on your system:", reprs).run()
            # no need to set an alternate selection if the job list not changed
            if len(reprs) != len(wanted_set):
                # wanted_set may have bad order, let's use it as a filter to
                # the original list
                chosen_jobs = [job for job in all_jobs if job in wanted_set]
                _logger.debug("remote: Selected jobs: %s", chosen_jobs)
                self.sa.modify_todo_list(chosen_jobs)
            manifest_repr = self.sa.get_manifest_repr()
            if manifest_repr:
                manifest_answers = ManifestBrowser(
                    "System Manifest:", manifest_repr).run()
                self.sa.save_manifest(manifest_answers)
        self.sa.finish_job_selection()
        self.run_jobs()

    def register_arguments(self, parser):
        parser.add_argument('host', help=_("target host"))
        parser.add_argument('launcher', nargs='?', help=_(
            "launcher definition file to use"))
        parser.add_argument('--port', type=int, default=18871, help=_(
            "port to connect to"))
        parser.add_argument('-u', '--user', help=_(
            "normal user to run non-root jobs"))

    def _handle_interrupt(self):
        """
        Returns True if the remote should keep running.
        And False if it should quit.
        """
        if self.launcher.ui_type == 'silent':
            self._sa.terminate()
            return False
        response = interrupt_dialog(self._target_host)
        if response == 'cancel':
            return True
        elif response == 'kill-master':
            return False
        elif response == 'kill-service':
            self._sa.terminate()
            return False
        elif response == 'abandon':
            self._sa.finalize_session()
            return True
        elif response == 'kill-command':
            self._sa.send_signal(signal.SIGKILL.value)
            return True

    def finish_session(self):
        print(self.C.header("Results"))
        if self.launcher.local_submission:
            # Disable SIGINT while we save local results
            with contextlib.ExitStack() as stack:
                tmp_sig = signal.signal(signal.SIGINT, signal.SIG_IGN)
                stack.callback(signal.signal, signal.SIGINT, tmp_sig)
                self._export_results()
        self.sa.finalize_session()
        return False

    def wait_and_continue(self):
        progress = self.sa.whats_up()[1]
        print("Rejoined session.")
        print("In progress: {} ({}/{})".format(
            progress[2], progress[0], progress[1]))
        self.wait_for_job()
        self.run_jobs()

    def _handle_last_job_after_resume(self, resumed_session_info):
        if self.launcher.ui_type == 'silent':
            time.sleep(20)
        else:
            resume_dialog(10)
        jobs_repr = json.loads(
            self.sa.get_jobs_repr([resumed_session_info['last_job']]))
        job = jobs_repr[-1]
        SimpleUI.header(job['name'])
        print(_("ID: {0}").format(job['id']))
        print(_("Category: {0}").format(job['category_name']))
        SimpleUI.horiz_line()
        print(
            _("Outcome") + ": " +
            SimpleUI.C.result(self.sa.get_job_result(job['id'])))

    def run_jobs(self, resumed_session_info=None):
        if resumed_session_info and resumed_session_info['last_job']:
            self._handle_last_job_after_resume(resumed_session_info)
        _logger.info("remote: Running jobs.")
        jobs = self.sa.get_session_progress()
        _logger.debug("remote: Jobs to be run:\n%s",
                      '\n'.join(['  ' + job for job in jobs]))
        total_num = len(jobs['done']) + len(jobs['todo'])

        jobs_repr = json.loads(
            self.sa.get_jobs_repr(jobs['todo'], len(jobs['done'])))

        self._run_jobs(jobs_repr, total_num)
        rerun_candidates = self.sa.get_rerun_candidates('manual')
        if rerun_candidates:
            if self.launcher.ui_type == 'interactive':
                while True:
                    if not self._maybe_manual_rerun_jobs():
                        break
        if self.launcher.auto_retry:
            while True:
                if not self._maybe_auto_rerun_jobs():
                    break
        self.finish_session()

    def resume_interacting(self, interaction):
        self.sa.remember_users_response('rollback')
        self.run_jobs()

    def wait_for_job(self, dont_finish=False):
        _logger.info("remote: Waiting for job to finish.")
        while True:
            state, payload = self.sa.monitor_job()
            if payload and not self._is_bootstrapping:
                for line in payload.splitlines():
                    if line.startswith('stderr'):
                        SimpleUI.red_text(line[6:])
                    elif line.startswith('stdout'):
                        SimpleUI.green_text(line[6:])
                    else:
                        SimpleUI.black_text(line[6:])
            if state == 'running':
                time.sleep(0.5)
                while True:
                    res = select.select([sys.stdin], [], [], 0)
                    if not res[0]:
                        break
                    # XXX: this assumes that sys.stdin is chunked in lines
                    buff = res[0][0].readline()
                    self.sa.transmit_input(buff)
                    if not buff:
                        break
            else:
                if dont_finish:
                    return
                self.finish_job()
                break

    def finish_job(self, result=None):
        _logger.info("remote: Finishing job with a result: %s", result)
        job_result = self.sa.finish_job(result)
        if not self._is_bootstrapping:
            SimpleUI.horiz_line()
            print(_("Outcome") + ": " + SimpleUI.C.result(job_result))

    def abandon(self):
        _logger.info("remote: Abandoning session.")
        self.sa.finalize_session()

    def restart(self):
        _logger.info("remote: Restarting session.")
        self.abandon()
        self.new_session()

    def local_export(self, exporter_id, transport, options=()):
        _logger.info("remote: Exporting locally'")
        rf = self.sa.cache_report(exporter_id, options)
        exported_stream = SpooledTemporaryFile(max_size=102400, mode='w+b')
        chunk_size = 16384
        with tqdm(
            total=rf.tell(),
            unit='B',
            unit_scale=True,
            unit_divisor=1024,
            disable=not self.is_interactive
        ) as pbar:
            rf.seek(0)
            while True:
                buf = rf.read(chunk_size)
                pbar.set_postfix(file=transport.url, refresh=False)
                pbar.update(chunk_size)
                if not buf:
                    break
                exported_stream.write(buf)
        exported_stream.seek(0)
        result = transport.send(exported_stream)
        return result

    def _maybe_auto_rerun_jobs(self):
        # create a list of jobs that qualify for rerunning
        rerun_candidates = self.sa.get_rerun_candidates('auto')
        # bail-out early if no job qualifies for rerunning
        if not rerun_candidates:
            return False
        # we wait before retrying
        delay = self.launcher.delay_before_retry
        _logger.info(_("Waiting {} seconds before retrying failed"
                       " jobs...".format(delay)))
        time.sleep(delay)
        # include resource jobs that jobs to retry depend on

        candidates = self.sa.prepare_rerun_candidates(rerun_candidates)
        self._run_jobs(
            json.loads(self.sa.get_jobs_repr(candidates)), len(candidates))
        return True

    def _maybe_manual_rerun_jobs(self):
        rerun_candidates = self.sa.get_rerun_candidates('manual')
        if not rerun_candidates:
            return False
        test_info_list = json.loads(
            self.sa.get_jobs_repr([j.id for j in rerun_candidates]))
        wanted_set = ReRunBrowser(
            _("Select jobs to re-run"), test_info_list, rerun_candidates).run()
        if not wanted_set:
            return False
        candidates = self.sa.prepare_rerun_candidates([
            job for job in rerun_candidates if job.id in wanted_set])
        self._run_jobs(
            json.loads(self.sa.get_jobs_repr(candidates)), len(candidates))
        return True

    def _run_jobs(self, jobs_repr, total_num=0):
        for job in jobs_repr:
            SimpleUI.header(
                _('Running job {} / {}').format(
                    job['num'], total_num,
                    fill='-'))
            SimpleUI.header(job['name'])
            print(_("ID: {0}").format(job['id']))
            print(_("Category: {0}").format(job['category_name']))
            SimpleUI.horiz_line()
            next_job = False
            while next_job is False:
                for interaction in self.sa.run_job(job['id']):
                    if interaction.kind == 'purpose':
                        SimpleUI.description(_('Purpose:'),
                                             interaction.message)
                    elif interaction.kind == 'description':
                        SimpleUI.description(_('Description:'),
                                             interaction.message)
                        if job['command'] is None:
                            cmd = 'run'
                        else:
                            cmd = SimpleUI(
                                None).wait_for_interaction_prompt(None)
                        if cmd == 'skip':
                            next_job = True
                        self.sa.remember_users_response(cmd)
                        self.wait_for_job(dont_finish=True)
                    elif interaction.kind in 'steps':
                        SimpleUI.description(_('Steps:'), interaction.message)
                        if job['command'] is None:
                            cmd = 'run'
                        else:
                            cmd = SimpleUI(
                                None).wait_for_interaction_prompt(None)
                        if cmd == 'skip':
                            next_job = True
                        self.sa.remember_users_response(cmd)
                    elif interaction.kind == 'verification':
                        self.wait_for_job(dont_finish=True)
                        if interaction.message:
                            SimpleUI.description(
                                _('Verification:'), interaction.message)
                        JobAdapter = namedtuple('job_adapter', ['command'])
                        job_lite = JobAdapter(job['command'])
                        try:
                            cmd = SimpleUI(None)._interaction_callback(
                                job_lite, interaction.extra._builder)
                            self.sa.remember_users_response(cmd)
                            self.finish_job(
                                interaction.extra._builder.get_result())
                            next_job = True
                            break
                        except ReRunJob:
                            next_job = False
                            self.sa.rerun_job(
                                job['id'],
                                interaction.extra._builder.get_result())
                            break
                    elif interaction.kind == 'comment':
                        new_comment = input(SimpleUI.C.BLUE(
                            _('Please enter your comments:') + '\n'))
                        self.sa.remember_users_response(new_comment + '\n')
                    elif interaction.kind == 'skip':
                        self.finish_job(
                            interaction.extra._builder.get_result())
                        next_job = True
                        break
                else:
                    self.wait_for_job()
                    break
            if next_job:
                continue



# This file is part of Checkbox.
#
# Copyright 2018-2019 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.
from plainbox.impl.secure.config import ValidationError
from plainbox.i18n import gettext as _

from checkbox_ng.config import load_configs


class CheckConfig():
    def invoked(self, ctx):
        config = load_configs()
        print(_("Configuration files:"))
        for filename in config.filename_list:
            print(" - {}".format(filename))
        for variable in config.Meta.variable_list:
            print("   [{0}]".format(variable.section))
            print("   {0}={1}".format(
                variable.name,
                variable.__get__(config, config.__class__)))
        for section in config.Meta.section_list:
            print("   [{0}]".format(section.name))
            section_value = section.__get__(config, config.__class__)
            if section_value:
                for key, value in sorted(section_value.items()):
                    print("   {0}={1}".format(key, value))
        if config.problem_list:
            print(_("Problems:"))
            for problem in config.problem_list:
                if isinstance(problem, ValidationError):
                    print(_(" - variable {0}: {1}").format(
                        problem.variable.name, problem.message))
                else:
                    print(" - {0}".format(problem.message))
            return 1
        else:
            print(_("No validation problems found"))
            return 0

    def register_arguments(self, parser):
        pass



# This file is part of Checkbox.
#
# Copyright 2018 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

import gettext
import logging
import os

from collections import OrderedDict

from plainbox.impl.providers import get_providers

import checkbox_ng

_ = gettext.gettext
_logger = logging.getLogger("checkbox-ng.launcher.subcommands")


def get_version_info():
    info = OrderedDict()
    if os.environ.get("SNAP"):
        return get_parts_info()
    else:
        return get_stack_version()


def get_stack_version():
    """
    Generate version information of Checkbox stack components as understood
    by Checkbox.
    """
    info = OrderedDict()
    info['checkbox-ng'] = checkbox_ng.__version__
    try:
        import checkbox_support
        support_ver = checkbox_support.__version__
    except AttributeError:
        import pkg_resources
        support_ver = pkg_resources.get_distribution(
            "checkbox-support").version
    except ModuleNotFoundError:
        support_ver = 'not available'
    info['checkbox-support'] = support_ver
    ignored_providers = [
        'com.canonical.plainbox:manifest',
        'com.canonical.plainbox:exporters',
        'com.canonical.plainbox:categories',
    ]
    # maybe filter-out the built-in providers (exporters, categories, etc.)
    for provider in get_providers():
        if provider.name not in ignored_providers:
            info[provider.name] = provider.version
    return info


def get_parts_info():
    """
    Generate information about the revisions of all the parts that were used
    to build the snap Checkbox is running from.
    """
    info = OrderedDict()
    parts_info_path = os.path.join(
        os.path.expandvars("$SNAP"), 'parts_meta_info')
    if not os.path.exists(parts_info_path):
        _logger.warning(_('Missing parts_meta_info file.'))
        return info
    with open(parts_info_path, 'rt') as f:
        content = f.read().splitlines()
    while(len(content) >= 4):
        part_name = content[0]
        # trim trailing ':'
        part_name = part_name[:-1]
        commit = content[1].split(' ')[0]
        content = content[4:]
        info[part_name] = commit
    if len(content) > 0:
        # XXX: soft-assert here that nothing's left to read
        _logger.warning(_('Found trailing data in parts_meta_info.'))
    return info



# This file is part of Checkbox.
#
# Copyright 2013-2019 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`checkbox_ng.config` -- CheckBoxNG configuration
=====================================================
"""
import gettext
import itertools
import logging
import os

from plainbox.impl.launcher import DefaultLauncherDefinition
from plainbox.impl.launcher import LauncherDefinition


_ = gettext.gettext

_logger = logging.getLogger("config")

def expand_all(path):
    return os.path.expandvars(os.path.expanduser(path))

def load_configs(launcher_file=None):
    # launcher can override the default name of config files to look for
    # so first we need to establish the filename to look for
    configs = []
    config_filename = 'checkbox.conf'
    launcher = DefaultLauncherDefinition()
    if launcher_file:
        configs.append(launcher_file)
        generic_launcher = LauncherDefinition()
        if not os.path.exists(launcher_file):
            _logger.error(_(
                "Unable to load launcher '%s'. File not found!"),
                launcher_file)
            raise SystemExit(1)
        generic_launcher.read(launcher_file)
        config_filename = os.path.expandvars(os.path.expanduser(
            generic_launcher.config_filename))
        launcher = generic_launcher.get_concrete_launcher()
    if os.path.isabs(config_filename):
        configs.append(config_filename)
    else:
        search_dirs = [
            '$SNAP_DATA',
            '/etc/xdg/',
            '~/.config/',
        ]
        for d in search_dirs:
            config = expand_all(os.path.join(d, config_filename))
            if os.path.exists(config):
                configs.append(config)
    launcher.read(configs)
    if launcher.problem_list:
        _logger.error(_("Unable to start launcher because of errors:"))
        for problem in launcher.problem_list:
            _logger.error("%s", str(problem))
        raise SystemExit(1)
    return launcher



# This file is part of Checkbox.
#
# Copyright 2015 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the session assistant module class."""

from plainbox.impl.secure.providers.v1 import Provider1
from plainbox.impl.session.assistant import SessionAssistant
from plainbox.impl.session.assistant import UsageExpectation
from plainbox.vendor import mock
from plainbox.vendor import morris


@mock.patch('plainbox.impl.session.assistant.get_providers')
class SessionAssistantTests(morris.SignalTestCase):

    """Tests for the SessionAssitant class."""

    APP_ID = 'app-id'
    APP_VERSION = '1.0'
    API_VERSION = '0.99'
    API_FLAGS = []

    def setUp(self):
        """Common set-up code."""
        self.sa = SessionAssistant(
            self.APP_ID, self.APP_VERSION, self.API_VERSION, self.API_FLAGS)
        # Monitor the provider_selected signal since some tests check it
        self.watchSignal(self.sa.provider_selected)
        # Create a few mocked providers that tests can use.
        # The all-important plainbox provider
        self.p1 = mock.Mock(spec_set=Provider1, name='p1')
        self.p1.namespace = 'com.canonical.plainbox'
        self.p1.name = 'com.canonical.plainbox:special'
        # An example 3rd party provider
        self.p2 = mock.Mock(spec_set=Provider1, name='p2')
        self.p2.namespace = 'pl.zygoon'
        self.p2.name = 'pl.zygoon:example'
        # A Canonical certification provider
        self.p3 = mock.Mock(spec_set=Provider1, name='p3')
        self.p3.namespace = 'com.canonical.certification'
        self.p3.name = 'com.canonical.certification:stuff'

    def _get_mock_providers(self):
        """Get some mocked provides for testing."""
        return [self.p1, self.p2, self.p3]

    def test_expected_call_sequence(self, mock_get_providers):
        """Track the sequence of allowed method calls."""
        mock_get_providers.return_value = self._get_mock_providers()
        # SessionAssistant.start_new_session() must now be allowed
        self.assertIn(self.sa.start_new_session,
                      UsageExpectation.of(self.sa).allowed_calls)
        # Call SessionAssistant.start_new_session()
        self.sa.start_new_session("just for testing")
        # SessionAssistant.start_new_session() must no longer allowed
        self.assertNotIn(self.sa.start_new_session,
                         UsageExpectation.of(self.sa).allowed_calls)
        # SessionAssistant.select_test_plan() must now be allowed
        self.assertIn(self.sa.select_test_plan,
                      UsageExpectation.of(self.sa).allowed_calls)
        # Use the manager to tidy up after the tests when normally you wouldnt
        # be allowed to
        self.sa._manager.destroy()



# This file is part of Checkbox.
#
# Copyright 2013-2021 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#   Daniel Manrique <roadmr@ubuntu.com>
#   Sylvain Pineau <sylvain.pineau@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
plainbox.impl.transport.test_certification
==========================================

Test definitions for plainbox.impl.certification module
"""

from io import BytesIO
from unittest import TestCase

from pkg_resources import resource_string
from plainbox.impl.transport import InvalidSecureIDError
from plainbox.impl.transport import TransportError
from plainbox.vendor import mock
from plainbox.vendor.mock import MagicMock
from requests.exceptions import ConnectionError, InvalidSchema, HTTPError
import requests

from checkbox_ng.certification import SubmissionServiceTransport


class SubmissionServiceTransportTests(TestCase):

    # URL are just here to exemplify, since we mock away all network access,
    # they're not really used.
    valid_url = "https://certification.canonical.com/submissions/submit"
    invalid_url = "htz://:3128"
    unreachable_url = "http://i.dont.exist"
    valid_secure_id = "a00D000000Kkk5j"
    valid_option_string = "secure_id={}".format(valid_secure_id)

    def setUp(self):
        self.sample_archive = BytesIO(resource_string(
            "plainbox", "test-data/tar-exporter/example-data.tar.xz"
        ))
        self.patcher = mock.patch('requests.post')
        self.mock_requests = self.patcher.start()

    def test_parameter_parsing(self):
        # Makes sense since I'm overriding the base class's constructor.
        transport = SubmissionServiceTransport(
            self.valid_url, self.valid_option_string)
        self.assertEqual(self.valid_url, transport.url)
        self.assertEqual(self.valid_secure_id,
                         transport.options['secure_id'])

    def test_invalid_length_secure_id_are_rejected(self):
        length = 14
        dummy_id = "a" * length
        option_string = "secure_id={}".format(dummy_id)
        with self.assertRaises(InvalidSecureIDError):
            SubmissionServiceTransport(self.valid_url, option_string)

    def test_invalid_characters_in_secure_id_are_rejected(self):
        option_string = "secure_id=aA0#"
        with self.assertRaises(InvalidSecureIDError):
            SubmissionServiceTransport(self.valid_url, option_string)

    def test_invalid_url(self):
        transport = SubmissionServiceTransport(
            self.invalid_url, self.valid_option_string)
        dummy_data = BytesIO(b"some data to send")
        requests.post.side_effect = InvalidSchema

        with self.assertRaises(TransportError):
            result = transport.send(dummy_data)
            self.assertIsNotNone(result)
        requests.post.assert_called_with(
            self.invalid_url, data=dummy_data)

    @mock.patch('checkbox_ng.certification.logger')
    def test_valid_url_cant_connect(self, mock_logger):
        transport = SubmissionServiceTransport(
            self.unreachable_url, self.valid_option_string)
        dummy_data = BytesIO(b"some data to send")
        requests.post.side_effect = ConnectionError
        with self.assertRaises(TransportError):
            result = transport.send(dummy_data)
            self.assertIsNotNone(result)
        requests.post.assert_called_with(self.unreachable_url,
                                         data=dummy_data)

    def test_send_success(self):
        transport = SubmissionServiceTransport(
            self.valid_url, self.valid_option_string)
        requests.post.return_value = MagicMock(name='response')
        requests.post.return_value.status_code = 200
        requests.post.return_value.text = '{"id": 768}'
        result = transport.send(self.sample_archive)
        self.assertTrue(result)

    def test_send_failure(self):
        transport = SubmissionServiceTransport(
            self.valid_url, self.valid_option_string)
        requests.post.return_value = MagicMock(name='response')
        requests.post.return_value.status_code = 412
        requests.post.return_value.text = 'Some error'
        # Oops, raise_for_status doesn't get fooled by my mocking,
        # so I have to mock *that* method as well..
        response = requests.Response()
        error = HTTPError(response=response)
        requests.post.return_value.raise_for_status = MagicMock(
            side_effect=error)
        with self.assertRaises(TransportError):
            transport.send(self.sample_archive)



# This file is part of Checkbox.
#
# Copyright 2015 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#   Sylvain Pineau <sylvain.pineau@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`plainbox.impl.exporter.jinja2` -- exporter using jinja2 templates
=======================================================================

.. warning::
    THIS MODULE DOES NOT HAVE A STABLE PUBLIC API
"""

import json
import re
from collections import OrderedDict
from datetime import datetime

from jinja2 import Environment
from jinja2 import FileSystemLoader
from jinja2 import Markup
from jinja2 import environmentfilter
from jinja2 import escape

from plainbox import get_version_string
from plainbox.abc import ISessionStateExporter
from plainbox.impl.exporter import SessionStateExporterBase
from plainbox.impl.result import OUTCOME_METADATA_MAP
from plainbox.impl.unit.exporter import ExporterError


#: Name-space prefix for Canonical Certification
CERTIFICATION_NS = 'com.canonical.certification::'


@environmentfilter
def do_strip_ns(_environment, unit_id, ns=CERTIFICATION_NS):
    """Remove the namespace part of the identifier."""
    # com.my.namespace::category/job-id → category/job-id
    rv = unit_id.split("::")[-1]
    rv = escape(rv)
    if _environment.autoescape:
        rv = Markup(rv)
    return rv


def do_is_name(text):
    """A filter for checking if something is equal to "name"."""
    return text == 'name'


def json_load_ordered_dict(text):
    """Render json dict in Jinja templates but keep keys ordering."""
    return json.loads(
        text, object_pairs_hook=OrderedDict)


def highlight_keys(text):
    """A filter for rendering keys as bold html text."""
    return re.sub('(\w+:\s)', r'<b>\1</b>', text)


class Jinja2SessionStateExporter(SessionStateExporterBase):

    """Session state exporter that renders output using jinja2 template."""

    supported_option_list = ('without-session-desc')

    def __init__(self, option_list=None, system_id="", timestamp=None,
                 client_version=None, client_name='plainbox',
                 exporter_unit=None):
        """
        Initialize a new Jinja2SessionStateExporter with given arguments.
        """
        super().__init__((), exporter_unit=exporter_unit)
        self._unit = exporter_unit
        self._system_id = system_id
        # Generate a time-stamp if needed
        self._timestamp = (
            timestamp or datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S"))
        # Use current version unless told otherwise
        self._client_version = client_version or get_version_string()
        # Remember client name
        self._client_name = client_name

        self.option_list = None
        self.template = None
        self.data = dict()
        paths = []
        if exporter_unit:
            self.data = exporter_unit.data
            # Add PROVIDER_DATA to the list of paths where to look for
            # templates
            paths.append(exporter_unit.data_dir)
        if "extra_paths" in self.data:
            paths.extend(self.data["extra_paths"])
        self.option_list = tuple(exporter_unit.option_list or ()) + tuple(
                option_list or ())
        loader = FileSystemLoader(paths)
        env = Environment(loader=loader, extensions=['jinja2.ext.autoescape'])
        self.customize_environment(env)

        def include_file(name):
            # This helper function insert static files literally into Jinja
            # templates without parsing them.
            return Markup(loader.get_source(env, name)[0])

        env.globals['include_file'] = include_file
        self.template = env.get_template(exporter_unit.template)

    @property
    def unit(self):
        """
        Exporter unit this exporter was created with.

        The exporter unit holds additional information that may be of use to
        applications, such as typical file name extension.
        """
        return self._unit

    def customize_environment(self, env):
        """Register filters and tests custom to the JSON exporter."""
        env.autoescape = True
        env.filters['jsonify'] = json.dumps
        env.filters['strip_ns'] = do_strip_ns
        env.filters['json_load_ordered_dict'] = json_load_ordered_dict
        env.filters['highlight_keys'] = highlight_keys
        env.tests['is_name'] = do_is_name

    def dump(self, data, stream):
        """
        Render report using jinja2 and dump it to stream.

        :param data:
            Dict to be used when rendering template instance
        :param stream:
            Byte stream to write to.

        """
        self.template.stream(data).dump(stream, encoding='utf-8')

    def dump_from_session_manager(self, session_manager, stream):
        """
        Extract data from session_manager and dump it into the stream.

        :param session_manager:
            SessionManager instance that manages session to be exported by
            this exporter
        :param stream:
            Byte stream to write to.

        """
        try:
            state = self._trim_session_manager(session_manager).state
            app_blob = state.metadata.app_blob
            app_blob_data = json.loads(app_blob.decode("UTF-8"))
        except ValueError:
            app_blob_data = {}
        data = {
            'OUTCOME_METADATA_MAP': OUTCOME_METADATA_MAP,
            'client_name': self._client_name,
            'client_version': self._client_version,
            'manager': session_manager,
            'app_blob': app_blob_data,
            'options': self.option_list,
            'system_id': self._system_id,
            'timestamp': self._timestamp,
        }
        data.update(self.data)
        self.dump(data, stream)
        self.validate(stream)

    def dump_from_session_manager_list(self, session_manager_list, stream):
        """
        Extract data from session_manager_list and dump them into the stream.

        :param session_manager_list:
            SessionManager instances that manages session to be exported by
            this exporter
        :param stream:
            Byte stream to write to.

        """
        data = {
            'OUTCOME_METADATA_MAP': OUTCOME_METADATA_MAP,
            'client_name': self._client_name,
            'client_version': self._client_version,
            'manager_list': session_manager_list,
            'app_blob': {},
            'options': self.option_list,
            'system_id': self._system_id,
            'timestamp': self._timestamp,
        }
        data.update(self.data)
        self.dump(data, stream)

    def get_session_data_subset(self, session_manager):
        """Compute a subset of session data."""
        return {
            'manager': session_manager,
            'options': self.option_list,
        }

    def validate(self, stream):
        # we need to validate the whole thing from the beginning
        pos = stream.tell()
        stream.seek(0)
        validator_fun = {
            'json': self.validate_json,
        }.get(self.unit.file_extension, lambda *_: [])
        problems = validator_fun(stream)
        # XXX: in case of problems we don't really need to .seek() back
        # but let's be safe
        stream.seek(pos)
        if problems:
            raise ExporterError(problems)

    def validate_json(self, stream):
        """
        Returns a list of things wrong that made the validation fail.
        """
        # keeping it as a method to make it tidy and consistent with
        # any other possible validator that may use self
        try:
            # manually reading the stream to ensure decoding
            raw = stream.read()
            s = raw.decode('utf-8') if type(raw) == bytes else raw
            json.loads(s)
            return []
        except Exception as exc:
            return [str(exc)]



# encoding: utf-8
# This file is part of Checkbox.
#
# Copyright 2012-2015 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
Definition of JobRunner class.

:mod:`plainbox.impl.runner` -- job runner
=========================================

.. warning::

    THIS MODULE DOES NOT HAVE STABLE PUBLIC API
"""

import collections
import contextlib
import datetime
import getpass
import gzip
import io
import logging
import os
import select
import string
import subprocess
import sys
import tempfile
import threading
import time


from plainbox.abc import IJobResult, IJobRunner
from plainbox.i18n import gettext as _
from plainbox.impl.result import IOLogRecord
from plainbox.impl.result import IOLogRecordWriter
from plainbox.impl.result import JobResultBuilder
from plainbox.impl.secure.config import Unset
from plainbox.vendor import extcmd
from plainbox.vendor import morris


logger = logging.getLogger("plainbox.runner")


def slugify(_string):
    """Transform any string to one that can be used in filenames."""
    valid_chars = frozenset(
        "-_.{}{}".format(string.ascii_letters, string.digits))
    return ''.join(c if c in valid_chars else '_' for c in _string)


class IOLogRecordGenerator(extcmd.DelegateBase):

    """Delegate for extcmd that generates io_log entries."""

    def on_begin(self, args, kwargs):
        """
        Internal method of extcmd.DelegateBase.

        Called when a command is being invoked.

        Begins tracking time (relative time entries)
        """
        self.last_msg = datetime.datetime.utcnow()

    def on_line(self, stream_name, line):
        """
        Internal method of extcmd.DelegateBase.

        Creates a new IOLogRecord and passes it to :meth:`on_new_record()`.
        Maintains a timestamp of the last message so that approximate delay
        between each piece of output can be recorded as well.
        """
        now = datetime.datetime.utcnow()
        delay = now - self.last_msg
        self.last_msg = now
        record = IOLogRecord(delay.total_seconds(), stream_name, line)
        self.on_new_record(record)

    @morris.signal
    def on_new_record(self, record):
        """
        Internal signal method of :class:`IOLogRecordGenerator`.

        Called when a new record is generated and needs to be processed.
        """
        # TRANSLATORS: io means input-output
        logger.debug(_("io log generated %r"), record)


class CommandOutputWriter(extcmd.DelegateBase):

    """
    Delegate for extcmd that writes output to a file on disk.

    The file itself is only opened once on_begin() gets called by extcmd. This
    makes it safe to instantiate this without worrying about dangling
    resources.
    """

    def __init__(self, stdout_path, stderr_path):
        """
        Initialize new writer.

        Just records output paths.
        """
        self.stdout_path = stdout_path
        self.stderr_path = stderr_path

    def on_begin(self, args, kwargs):
        """
        Internal method of extcmd.DelegateBase.

        Called when a command is being invoked
        """
        self.stdout = open(self.stdout_path, "wb")
        self.stderr = open(self.stderr_path, "wb")

    def on_end(self, returncode):
        """
        Internal method of extcmd.DelegateBase.

        Called when a command finishes running
        """
        self.stdout.close()
        self.stderr.close()

    def on_abnormal_end(self, signal_num):
        """
        Internal method of extcmd.DelegateBase.

        Called when a command abnormally finishes running
        """
        self.stdout.close()
        self.stderr.close()

    def on_line(self, stream_name, line):
        """
        Internal method of extcmd.DelegateBase.

        Called for each line of output.
        """
        if stream_name == 'stdout':
            self.stdout.write(line)
        elif stream_name == 'stderr':
            self.stderr.write(line)


class FallbackCommandOutputPrinter(extcmd.DelegateBase):

    """
    Delegate for extcmd that prints all output to stdout.

    This delegate is only used as a fallback when no delegate was explicitly
    provided to a JobRunner instance.
    """

    def __init__(self, prompt):
        """Initialize a new fallback command output printer."""
        self._prompt = prompt
        self._lineno = collections.defaultdict(int)
        self._abort = False

    def on_line(self, stream_name, line):
        """
        Internal method of extcmd.DelegateBase.

        Called for each line of output. Normally each line is just printed
        (assuming UTF-8 encoding) If decoding fails for any reason that and all
        subsequent lines are ignored.
        """
        if self._abort:
            return
        self._lineno[stream_name] += 1
        try:
            print("(job {}, <{}:{:05}>) {}".format(
                self._prompt, stream_name, self._lineno[stream_name],
                line.decode('UTF-8').rstrip()))
        except UnicodeDecodeError:
            self._abort = True


class JobRunnerUIDelegate(extcmd.DelegateBase):

    """
    Delegate for extcmd that delegates extcmd events to IJobRunnerUI.

    The file itself is only opened once on_begin() gets called by extcmd. This
    makes it safe to instantiate this without worrying about dangling
    resources.

    The instance attribute 'ui' can be changed at any time. It can also be set
    to None to silence all notifications from execution progress of external
    programs.
    """

    def __init__(self, ui=None):
        """
        Initialize the JobRunnerUIDelegate.

        :param ui:
            (optional) an instance of IJobRunnerUI to delegate events to
        """
        self.ui = ui

    def on_begin(self, args, kwargs):
        """
        Internal method of extcmd.DelegateBase.

        Called when a command is being invoked
        """
        if self.ui is not None:
            self.ui.about_to_execute_program(args, kwargs)

    def on_end(self, returncode):
        """
        Internal method of extcmd.DelegateBase.

        Called when a command finishes running
        """
        if self.ui is not None:
            self.ui.finished_executing_program(returncode)

    def on_abnormal_end(self, signal_num):
        """
        Internal method of extcmd.DelegateBase.

        Called when a command abnormally finishes running

        The negated signal number is used as the exit code of the program and
        fed into the UI (if any)
        """
        if self.ui is not None:
            self.ui.finished_executing_program(-signal_num)

    def on_line(self, stream_name, line):
        """
        Internal method of extcmd.DelegateBase.

        Called for each line of output.
        """
        if self.ui is not None:
            self.ui.got_program_output(stream_name, line)

    def on_chunk(self, stream_name, chunk):
        """
        Internal method of extcmd.DelegateBase.

        Called for each chunk of output.
        """
        if self.ui is not None:
            self.ui.got_program_output(stream_name, chunk)



# This file is part of Checkbox.
#
# Copyright 2012-2018 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

import collections
import datetime
import logging
import sys

from plainbox.abc import IJobRunnerUI
from plainbox.i18n import gettext as _
from plainbox.impl.color import Colorizer


logger = logging.getLogger("checkbox-ng.launcher.run")


Action = collections.namedtuple("Action", "accel label cmd")


class ActionUI:
    """
    A simple user interface to display a list of actions and let the user to
    pick one
    """

    def __init__(self, action_list, prompt=None, color=None):
        """
        :param action_list:
            A list of 3-tuples (accel, label, cmd)
        :prompt:
            An optional prompt string
        :returns:
            cmd of the selected action or None
        """
        if prompt is None:
            prompt = _("Pick an action")
        self.action_list = action_list
        self.prompt = prompt
        self.C = Colorizer(color)

    def run(self):
        long_hint = "\n".join(
            "  {accel} => {label}".format(
                accel=self.C.BLUE(action.accel) if action.accel else ' ',
                label=action.label)
            for action in self.action_list)
        short_hint = ''.join(action.accel for action in self.action_list)
        while True:
            try:
                print(self.C.BLUE(self.prompt))
                print(long_hint)
                choice = input("[{}]: ".format(self.C.BLUE(short_hint)))
            except EOFError:
                return None
            else:
                for action in self.action_list:
                    if choice == action.accel or choice == action.label:
                        return action.cmd


class SilentUI(IJobRunnerUI):

    def considering_job(self, job, job_state):
        pass

    def about_to_start_running(self, job, job_state):
        pass

    def wait_for_interaction_prompt(self, job):
        pass

    def started_running(self, job, job_state):
        pass

    def about_to_execute_program(self, args, kwargs):
        pass

    def finished_executing_program(self, returncode):
        pass

    def got_program_output(self, stream_name, line):
        pass

    def finished_running(self, job, job_state, job_result):
        pass

    def notify_about_description(self, job):
        pass

    def notify_about_purpose(self, job):
        pass

    def notify_about_steps(self, job):
        pass

    def notify_about_verification(self, job):
        pass

    def job_cannot_start(self, job, job_state, job_result):
        pass

    def finished(self, job, job_state, job_result):
        pass

    def pick_action_cmd(self, action_list, prompt=None):
        pass

    def noreturn_job(self):
        pass


class NormalUI(IJobRunnerUI):

    STREAM_MAP = {
        'stdout': sys.stdout,
        'stderr': sys.stderr
    }

    def __init__(self, color, show_cmd_output=True):
        self.show_cmd_output = show_cmd_output
        self.C = Colorizer(color)
        self._color = color

    def considering_job(self, job, job_state):
        print(self.C.header(job.tr_summary(), fill='-'))
        print(_("ID: {0}").format(job.id))
        print(_("Category: {0}").format(job_state.effective_category_id))

    def about_to_start_running(self, job, job_state):
        pass

    def wait_for_interaction_prompt(self, job):
        return self.pick_action_cmd([
            Action('', _("press ENTER to continue"), 'run'),
            Action('c', _('add a comment'), 'comment'),
            Action('s', _("skip this job"), 'skip'),
            Action('q', _("save the session and quit"), 'quit')
        ])

    def started_running(self, job, job_state):
        pass

    def about_to_execute_program(self, args, kwargs):
        if self.show_cmd_output:
            print(self.C.BLACK("... 8< -".ljust(80, '-')))
        else:
            print(self.C.BLACK("(" + _("Command output hidden") + ")"))

    def got_program_output(self, stream_name, line):
        if not self.show_cmd_output:
            return
        stream = self.STREAM_MAP[stream_name]
        stream = {
            'stdout': sys.stdout,
            'stderr': sys.stderr
        }[stream_name]
        try:
            if stream_name == 'stdout':
                print(self.C.GREEN(line.decode("UTF-8")),
                      end='', file=stream)
            elif stream_name == 'stderr':
                print(self.C.RED(line.decode("UTF-8")),
                      end='', file=stream)
        except UnicodeDecodeError:
            self.show_cmd_output = False
            print(self.C.BLACK("(" + _("Hiding binary test output") + ")"))
        stream.flush()

    def finished_executing_program(self, returncode):
        if self.show_cmd_output:
            print(self.C.BLACK("- >8 ---".rjust(80, '-')))

    def finished_running(self, job, state, result):
        pass

    def notify_about_description(self, job):
        if job.tr_description() is not None:
            print(self.C.CYAN(job.tr_description()))

    def notify_about_purpose(self, job):
        if job.tr_purpose() is not None:
            print(self.C.WHITE(_("Purpose:")))
            print()
            print(self.C.CYAN(job.tr_purpose()))
            print()
        else:
            self.notify_about_description(job)

    def notify_about_steps(self, job):
        if job.tr_steps() is not None:
            print(self.C.WHITE(_("Steps:")))
            print()
            print(self.C.CYAN(job.tr_steps()))
            print()

    def notify_about_verification(self, job):
        if job.tr_verification() is not None:
            print(self.C.WHITE(_("Verification:")))
            print()
            print(self.C.CYAN(job.tr_verification()))
            print()

    def job_cannot_start(self, job, job_state, result):
        print(_("Job cannot be started because:"))
        for inhibitor in job_state.readiness_inhibitor_list:
            print(" - {}".format(self.C.YELLOW(inhibitor)))

    def finished(self, job, job_state, result):
        self._print_result_outcome(result)

    def _print_result_outcome(self, result):
        print(_("Outcome") + ": " + self.C.result(result))

    def pick_action_cmd(self, action_list, prompt=None):
        return ActionUI(action_list, prompt, self._color).run()

    def noreturn_job(self):
        print(self.C.RED(_("Waiting for the system to shut down or"
                           " reboot...")))


class ReRunJob(Exception):
    """
    Exception raised from _interaction_callback to indicate that a job should
    be re-started.
    """


def seconds_to_human_duration(seconds: float) -> str:
    """ Convert ammount of seconds to human readable duration string. """
    delta = datetime.timedelta(seconds=round(seconds))
    return str(delta)



# This file is part of Checkbox.
#
# Copyright 2018 Canonical Ltd.
# Authors:
#     Sylvain Pineau <sylvain.pineau@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`checkbox-ng.launcher.merge_reports` -- merge-reports sub-command
======================================================================
"""
import json
import os
import tarfile
from tempfile import TemporaryDirectory

from plainbox.impl.ctrl import gen_rfc822_records_from_io_log
from plainbox.impl.providers.special import get_exporters
from plainbox.impl.resource import Resource
from plainbox.impl.result import IOLogRecord
from plainbox.impl.result import MemoryJobResult
from plainbox.impl.session import SessionManager
from plainbox.impl.unit.category import CategoryUnit
from plainbox.impl.unit.job import JobDefinition


#: Name-space prefix for Canonical Certification
CERTIFICATION_NS = 'com.canonical.certification::'


class MergeReports():
    def register_arguments(self, parser):
        parser.add_argument(
            'submission', nargs='*', metavar='SUBMISSION',
            help='submission tarball')
        parser.add_argument(
            '-o', '--output-file', metavar='FILE', required=True,
            help='save combined test results to the specified FILE')

    def _parse_submission(self, submission, tmpdir, mode="list"):
        try:
            with tarfile.open(submission) as tar:
                tar.extractall(tmpdir.name)
                with open(os.path.join(
                          tmpdir.name, 'submission.json')) as f:
                    data = json.load(f)
            for result in data['results']:
                result['plugin'] = 'shell'  # Required so default to shell
                result['summary'] = result['name']
                # 'id' field in json file only contains partial id
                result['id'] = result.get('full_id', result['id'])
                if "::" not in result['id']:
                    result['id'] = CERTIFICATION_NS + result['id']
                if mode == "list":
                    self.job_list.append(JobDefinition(result))
                elif mode == "dict":
                    self.job_dict[result['id']] = JobDefinition(result)
            for result in data['resource-results']:
                result['plugin'] = 'resource'
                result['summary'] = result['name']
                # 'id' field in json file only contains partial id
                result['id'] = result.get('full_id', result['id'])
                if "::" not in result['id']:
                    result['id'] = CERTIFICATION_NS + result['id']
                if mode == "list":
                    self.job_list.append(JobDefinition(result))
                elif mode == "dict":
                    self.job_dict[result['id']] = JobDefinition(result)
            for result in data['attachment-results']:
                result['plugin'] = 'attachment'
                result['summary'] = result['name']
                # 'id' field in json file only contains partial id
                result['id'] = result.get('full_id', result['id'])
                if "::" not in result['id']:
                    result['id'] = CERTIFICATION_NS + result['id']
                if mode == "list":
                    self.job_list.append(JobDefinition(result))
                elif mode == "dict":
                    self.job_dict[result['id']] = JobDefinition(result)
            for cat_id, cat_name in data['category_map'].items():
                if mode == "list":
                    self.category_list.append(
                        CategoryUnit({'id': cat_id, 'name': cat_name}))
                elif mode == "dict":
                    self.category_dict[cat_id] = CategoryUnit(
                        {'id': cat_id, 'name': cat_name})
        except OSError as e:
            raise SystemExit(e)
        except KeyError as e:
            self._output_potential_action(str(e))
            raise SystemExit(e)
        return data['title']

    def _populate_session_state(self, job, state):
        io_log = [
            IOLogRecord(count, 'stdout', line.encode('utf-8'))
            for count, line in enumerate(
                job.get_record_value('io_log').splitlines(
                    keepends=True))
        ]
        result = MemoryJobResult({
            'outcome': job.get_record_value('outcome',
                                            job.get_record_value('status')),
            'comments': job.get_record_value('comments'),
            'execution_duration': job.get_record_value('duration'),
            'io_log': io_log,
        })
        state.update_job_result(job, result)
        if job.plugin == 'resource':
            new_resource_list = []
            for record in gen_rfc822_records_from_io_log(job, result):
                resource = Resource(record.data)
                new_resource_list.append(resource)
            if not new_resource_list:
                new_resource_list = [Resource({})]
            state.set_resource_list(job.id, new_resource_list)
        job_state = state.job_state_map[job.id]
        job_state.effective_category_id = job.get_record_value(
            'category_id', 'com.canonical.plainbox::uncategorised')
        job_state.effective_certification_status = job.get_record_value(
            'certification_status', 'unspecified')

    def _create_exporter(self, exporter_id):
        exporter_map = {}
        exporter_units = get_exporters().unit_list
        for unit in exporter_units:
            if unit.Meta.name == 'exporter':
                support = unit.support
                if support:
                    exporter_map[unit.id] = support
        exporter_support = exporter_map[exporter_id]
        return exporter_support.exporter_cls(
            [], exporter_unit=exporter_support)

    def _output_potential_action(self, message):
        hint = ""
        keys = ['resource', 'attachment']
        for key in keys:
            if key in message:
                hint = ("Make sure your input submission provides {}-related "
                        "information.".format(key))
        if hint:
            print("Fail to merge. " + hint)
        else:
            print("Fail to merge.")

    def invoked(self, ctx):
        manager_list = []
        for submission in ctx.args.submission:
            tmpdir = TemporaryDirectory()
            self.job_list = []
            self.category_list = []
            session_title = self._parse_submission(submission, tmpdir)
            manager = SessionManager.create_with_unit_list(
                self.job_list + self.category_list)
            manager.state.metadata.title = session_title
            for job in self.job_list:
                self._populate_session_state(job, manager.state)
            manager_list.append(manager)
        exporter = self._create_exporter(
            'com.canonical.plainbox::html-multi-page')
        with open(ctx.args.output_file, 'wb') as stream:
            exporter.dump_from_session_manager_list(manager_list, stream)
        print(ctx.args.output_file)



# This file is part of Checkbox.
#
# Copyright 2018 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.
import fnmatch
import io
import json
import gettext
import logging
import os
import pwd
import time
from collections import namedtuple
from contextlib import suppress
from tempfile import SpooledTemporaryFile
from threading import Thread, Lock

from plainbox.impl.execution import UnifiedRunner
from plainbox.impl.session.assistant import SessionAssistant
from plainbox.impl.session.assistant import SA_RESTARTABLE
from plainbox.impl.session.jobs import InhibitionCause
from plainbox.impl.session.storage import WellKnownDirsHelper
from plainbox.impl.secure.sudo_broker import is_passwordless_sudo
from plainbox.impl.result import JobResultBuilder
from plainbox.impl.result import MemoryJobResult
from plainbox.abc import IJobResult

from checkbox_ng.config import load_configs
from checkbox_ng.launcher.run import SilentUI

import psutil

_ = gettext.gettext

_logger = logging.getLogger("plainbox.session.remote_assistant")

Interaction = namedtuple('Interaction', ['kind', 'message', 'extra'])


class Interaction(namedtuple('Interaction', ['kind', 'message', 'extra'])):
    __slots__ = ()

    def __new__(cls, kind, message="", extra=None):
        return super(Interaction, cls).__new__(cls, kind, message, extra)


Idle = 'idle'
Started = 'started'
Bootstrapping = 'bootstrapping'
Bootstrapped = 'bootstrapped'
TestsSelected = 'testsselected'
Running = 'running'
Interacting = 'interacting'
Finalizing = 'finalizing'


class BufferedUI(SilentUI):
    """UI type that queues the output for later reading."""

    def __init__(self):
        super().__init__()
        self.lock = Lock()
        self._output = io.StringIO()

    def _ignore_program_output(self, stream_name, line):
        pass

    def got_program_output(self, stream_name, line):
        with self.lock:
            try:
                self._output.write(stream_name + line.decode("UTF-8"))
            except UnicodeDecodeError:
                # Don't start a slave->master transfer for binary attachments
                self._output.write("hidden(Hiding binary test output)\n")
                self.got_program_output = self._ignore_program_output

    def get_output(self):
        """Returns all the output queued up since previous call."""
        with self.lock:
            output = self._output.getvalue()
            self._output = io.StringIO()
            return output


class RemoteSilentUI(SilentUI):
    """SilentUI + fake get_output."""

    def __init__(self):
        super().__init__()
        self._msg = "hidden(Command output hidden)"

    def get_output(self):
        msg = self._msg
        self._msg = ''
        return msg


class BackgroundExecutor(Thread):
    def __init__(self, sa, job_id, real_run, ui=RemoteSilentUI()):
        super().__init__()
        self._sa = sa
        self._job_id = job_id
        self._real_run = real_run
        self._ui = ui
        self._builder = None
        self._started_real_run = False
        self._sa.session_change_lock.acquire()
        self.start()
        _logger.debug("BackgroundExecutor started for %s" % job_id)

    def wait(self):
        while self.is_alive() or not self._started_real_run:
            # return control to RPC server
            time.sleep(0.1)
        self.join()
        return self._builder

    def run(self):
        self._started_real_run = True
        self._builder = self._real_run(self._job_id, self._ui, False)
        _logger.debug("Finished running")

    def outcome(self):
        return self._builder.outcome


class RemoteSessionAssistant():
    """Remote execution enabling wrapper for the SessionAssistant"""

    REMOTE_API_VERSION = 11

    def __init__(self, cmd_callback):
        _logger.debug("__init__()")
        self._cmd_callback = cmd_callback
        self._session_change_lock = Lock()
        self._operator_lock = Lock()
        self._ui = BufferedUI()
        self._input_piping = os.pipe()
        self._passwordless_sudo = is_passwordless_sudo()
        self.terminate_cb = None
        self._pipe_from_master = open(self._input_piping[1], 'w')
        self._pipe_to_subproc = open(self._input_piping[0])
        self._reset_sa()
        self._currently_running_job = None

    def _reset_sa(self):
        _logger.info("Resetting RSA")
        self._state = Idle
        self._sa = SessionAssistant('service', api_flags={SA_RESTARTABLE})
        self._be = None
        self._session_id = ""
        self._jobs_count = 0
        self._job_index = 0
        self._currently_running_job = None  # XXX: yuck!
        self._last_job = None
        self._current_comments = ""
        self._last_response = None
        self._normal_user = ''
        self.session_change_lock.acquire(blocking=False)
        self.session_change_lock.release()

    @property
    def session_change_lock(self):
        return self._session_change_lock

    @property
    def config(self):
        return self._sa.config

    def update_app_blob(self, app_blob):
        self._sa.update_app_blob(app_blob)

    def allowed_when(*states):
        def wrap(f):
            def fun(self, *args):
                if self._state not in states:
                    raise AssertionError(
                        "expected %s, is %s" % (states, self._state))
                return f(self, *args)
            return fun
        return wrap

    def interact(self, interaction):
        self._state = Interacting
        self._current_interaction = interaction
        yield self._current_interaction

    @allowed_when(Interacting)
    def remember_users_response(self, response):
        if response == 'rollback':
            self._currently_running_job = None
            self.session_change_lock.acquire(blocking=False)
            self.session_change_lock.release()
            self._current_comments = ""
            self._state = TestsSelected
            return
        self._last_response = response
        self._state = Running

    def _set_envvar_from_proc(self, name):
        for path in os.listdir('/proc/'):
            with suppress(Exception):
                env = open(
                    os.path.join('/proc/', path, 'environ')).read().split('\0')
                for key, val in [item.split('=') for item in env]:
                    if key == name:
                        return val
        return ''

    def _prepare_display_without_psutil(self):
        uid = pwd.getpwnam(self._normal_user).pw_uid
        return {
            'DISPLAY': self._set_envvar_from_proc('DISPLAY'),
            'XAUTHORITY': self._set_envvar_from_proc('XAUTHORITY'),
            'XDG_SESSION_TYPE': self._set_envvar_from_proc('XDG_SESSION_TYPE'),
            'XDG_RUNTIME_DIR': '/run/user/{}'.format(uid),
            'DBUS_SESSION_BUS_ADDRESS':
                'unix:path=/run/user/{}/bus'.format(uid)
        }

    def prepare_extra_env(self):
        # If possible also set the DISPLAY env var
        # i.e when a user desktop session is running
        for p in psutil.pids():
            try:
                p_environ = psutil.Process(p).environ()
                p_user = psutil.Process(p).username()
            except psutil.AccessDenied:
                continue
            except AttributeError:
                # psutil < 4.0.0 doesn't provide Process.environ()
                return self._prepare_display_without_psutil()
            except psutil.NoSuchProcess:
                # quietly ignore the process that died before we had a chance
                # to read the environment from them
                continue
            if (
                "DISPLAY" in p_environ and
                "XAUTHORITY" in p_environ and
                p_user != 'gdm'
            ):  # gdm uses :1024
                uid = pwd.getpwnam(self._normal_user).pw_uid
                return {
                    'DISPLAY': p_environ['DISPLAY'],
                    'XAUTHORITY': p_environ['XAUTHORITY'],
                    'XDG_SESSION_TYPE': p_environ['XDG_SESSION_TYPE'],
                    'XDG_RUNTIME_DIR': '/run/user/{}'.format(uid),
                    'DBUS_SESSION_BUS_ADDRESS':
                        'unix:path=/run/user/{}/bus'.format(uid)
                }
        return {}

    @allowed_when(Idle)
    def start_session(self, configuration):
        self._reset_sa()
        _logger.info("start_session: %r", configuration)
        session_title = 'remote'
        session_desc = 'remote session'
        session_type = 'remote'

        self._launcher = load_configs()
        if configuration['launcher']:
            self._launcher.read_string(configuration['launcher'], False)
            if self._launcher.session_title:
                session_title = self._launcher.session_title
            if self._launcher.session_desc:
                session_desc = self._launcher.session_desc

        self._sa.use_alternate_configuration(self._launcher)

        if configuration['normal_user']:
            self._normal_user = configuration['normal_user']
        else:
            self._normal_user = self._launcher.normal_user
            if not self._normal_user:
                self._normal_user = _guess_normal_user()
        runner_kwargs = {
            'normal_user_provider': lambda: self._normal_user,
            'stdin': self._pipe_to_subproc,
            'extra_env': self.prepare_extra_env,
        }
        self._sa.start_new_session(session_title, UnifiedRunner, runner_kwargs)
        new_blob = json.dumps({
            'description': session_desc,
            'type': session_type,
            'launcher': configuration['launcher'],
            'effective_normal_user': self._normal_user,
        }).encode("UTF-8")
        self._sa.update_app_blob(new_blob)
        self._sa.configure_application_restart(self._cmd_callback)

        self._session_id = self._sa.get_session_id()
        tps = self._sa.get_test_plans()
        filtered_tps = set()
        for filter in self._launcher.test_plan_filters:
            filtered_tps.update(fnmatch.filter(tps, filter))
        filtered_tps = list(filtered_tps)
        response = zip(filtered_tps, [self._sa.get_test_plan(
            tp).name for tp in filtered_tps])
        self._state = Started
        self._available_testplans = sorted(
            response, key=lambda x: x[1])  # sorted by name
        return self._available_testplans

    @allowed_when(Started)
    def prepare_bootstrapping(self, test_plan_id):
        """Save picked test plan to the app blob."""
        _logger.debug("prepare_bootstrapping: %r", test_plan_id)
        self._sa.update_app_blob(json.dumps(
            {'testplan_id': test_plan_id, }).encode("UTF-8"))
        self._sa.select_test_plan(test_plan_id)
        # TODO: REMOTE API RAPI: Change this API on the next RAPI bump
        # previously the function returned bool signifying the need for sudo
        # password. With slave being guaranteed to never need it anymor
        # we can make this funciton return nothing
        return False

    @allowed_when(Started)
    def get_bootstrapping_todo_list(self):
        return self._sa.get_bootstrap_todo_list()

    def finish_bootstrap(self):
        self._sa.finish_bootstrap()
        self._state = Bootstrapped
        if self._launcher.auto_retry:
            for job_id in self._sa.get_static_todo_list():
                job_state = self._sa.get_job_state(job_id)
                job_state.attempts = self._launcher.max_attempts
        return self._sa.get_static_todo_list()

    def get_manifest_repr(self):
        return self._sa.get_manifest_repr()

    def save_manifest(self, manifest_answers):
        return self._sa.save_manifest(manifest_answers)

    def modify_todo_list(self, chosen_jobs):
        self._sa.use_alternate_selection(chosen_jobs)

    def finish_job_selection(self):
        self._jobs_count = len(self._sa.get_dynamic_todo_list())
        self._state = TestsSelected

    @allowed_when(Interacting)
    def rerun_job(self, job_id, result):
        self._sa.use_job_result(job_id, result)
        self.session_change_lock.acquire(blocking=False)
        self.session_change_lock.release()
        self._state = TestsSelected

    def _get_ui_for_job(self, job):
        show_out = True
        if self._launcher.output == 'hide-resource-and-attachment':
            if job.plugin in ('local', 'resource', 'attachment'):
                show_out = False
        elif self._launcher.output in ['hide', 'hide-automated']:
            if job.plugin in ('shell', 'local', 'resource', 'attachment'):
                show_out = False
        if 'suppress-output' in job.get_flag_set():
            show_out = False
        if show_out:
            self._ui = BufferedUI()
        else:
            self._ui = RemoteSilentUI()
        return self._ui

    @allowed_when(TestsSelected)
    def run_job(self, job_id):
        """
        Depending on the type of the job, run_job can yield different number
        of Interaction instances.
        """
        _logger.debug("run_job: %r", job_id)
        self._job_index = self._jobs_count - len(
            self._sa.get_dynamic_todo_list()) + 1
        self._currently_running_job = job_id
        self._current_comments = ""
        job = self._sa.get_job(job_id)
        job_state = self._sa.get_job_state(job_id)

        if not job_state.can_start():
            outcome = IJobResult.OUTCOME_NOT_SUPPORTED
            for inhibitor in job_state.readiness_inhibitor_list:
                if (
                    inhibitor.cause == InhibitionCause.FAILED_RESOURCE and
                    'fail-on-resource' in job.get_flag_set()
                ):
                    outcome = IJobResult.OUTCOME_FAIL
                    break
                elif inhibitor.cause != InhibitionCause.FAILED_DEP:
                    continue
                related_job_state = self._sa._context.state.job_state_map[
                    inhibitor.related_job.id]
                if related_job_state.result.outcome == IJobResult.OUTCOME_SKIP:
                    outcome = IJobResult.OUTCOME_SKIP

            def cant_start_builder(*args, **kwargs):
                result_builder = JobResultBuilder(
                    outcome=outcome,
                    comments=job_state.get_readiness_description())
                return result_builder
            self._be = BackgroundExecutor(self, job_id, cant_start_builder)
            yield from self.interact(Interaction('skip', None, self._be))

        if job.plugin in [
                'manual', 'user-interact-verify', 'user-interact']:
            may_comment = True
            while may_comment:
                may_comment = False
                if job.tr_description() and not job.tr_purpose():
                    yield from self.interact(
                        Interaction('description', job.tr_description()))
                if job.tr_purpose():
                    yield from self.interact(
                        Interaction('purpose', job.tr_purpose()))
                if job.tr_steps():
                    yield from self.interact(
                        Interaction('steps', job.tr_steps()))
                if self._last_response == 'comment':
                    yield from self.interact(Interaction('comment'))
                    if self._last_response:
                        self._current_comments += self._last_response
                    may_comment = True
                    continue
            if self._last_response == 'skip':
                def skipped_builder(*args, **kwargs):
                    result_builder = JobResultBuilder(
                        outcome=IJobResult.OUTCOME_SKIP,
                        comments=_("Explicitly skipped before execution"))
                    if self._current_comments != "":
                        result_builder.comments = self._current_comments
                    return result_builder
                self._be = BackgroundExecutor(
                    self, job_id, skipped_builder)
                yield from self.interact(
                    Interaction('skip', job.verification, self._be))
        if job.command:
            self._state = Running
            ui = self._get_ui_for_job(job)
            self._be = BackgroundExecutor(self, job_id, self._sa.run_job, ui)
        else:
            def undecided_builder(*args, **kwargs):
                return JobResultBuilder(outcome=IJobResult.OUTCOME_UNDECIDED)
            self._be = BackgroundExecutor(self, job_id, undecided_builder)
        if self._sa.get_job(self._currently_running_job).plugin in [
                'manual', 'user-interact-verify']:
            yield from self.interact(
                Interaction('verification', job.verification, self._be))

    @allowed_when(Started, Bootstrapping)
    def run_bootstrapping_job(self, job_id):
        self._currently_running_job = job_id
        self._state = Bootstrapping
        self._be = BackgroundExecutor(self, job_id, self._sa.run_job)

    @allowed_when(Running, Bootstrapping, Interacting, TestsSelected)
    def monitor_job(self):
        """
        Check the state of the currently running job.

        :returns:
            (state, payload) tuple.
            Payload conveys detailed info that's characteristic
            to the current state.
        """
        _logger.debug("monitor_job()")
        # either return [done, running, awaiting response]
        # TODO: handle awaiting_response (reading from stdin by the job)
        if self._be and self._be.is_alive():
            return ('running', self._ui.get_output())
        else:
            return ('done', self._ui.get_output())

    def get_remote_api_version(self):
        return self.REMOTE_API_VERSION

    def whats_up(self):
        """
        Check what is remote-service up to
        :returns:
            (state, payload) tuple.
        """
        _logger.debug("whats_up() -> %r", self._state)
        payload = None
        if self._state == Running:
            payload = (
                self._job_index, self._jobs_count, self._currently_running_job
            )
        if self._state == TestsSelected and not self._currently_running_job:
            payload = {'last_job': self._last_job}
        elif self._state == Started:
            payload = self._available_testplans
        elif self._state == Interacting:
            payload = self._current_interaction
        elif self._state == Bootstrapped:
            payload = self._sa.get_static_todo_list()
        return self._state, payload

    def terminate(self):
        if self.terminate_cb:
            self.terminate_cb()

    def get_session_progress(self):
        """Return list of completed and not completed jobs in a dict."""

        _logger.debug("get_session_progress()")
        return {
            "done": self._sa.get_dynamic_done_list(),
            "todo": self._sa.get_dynamic_todo_list(),
        }

    def get_master_public_key(self):
        # TODO: REMOTE API RAPI: Remove this API on the next RAPI bump
        # this key is only for RAPI compliance. It will never be used as
        # this master requires slave to be completely sudoless
        return (
            b'-----BEGIN PUBLIC KEY-----\nMIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMII'
            b'BCgKCAQEA5r0bjOA+IH5lDKkW3OYb\nDuEjf5VKgUlDSJJuyBlfLTBIXZ8j3s98'
            b'6AbV0zB62rAcgiFrBOzx51IzBDBmHI8V\nYYpEa+q4OP4yprYpSg6xzX6LRQapC'
            b'Iv9BAqN4MWrKBukGMzJyemIVEPv4BSHL5L/\nLY98Mwh4dAXxj5ZdsoVPqgeMo8'
            b'dxfYEOwVRJvSkseIhxRL6tvgP37c48ApUyjdUO\n3C2YgqJRx7mKKDyLOvhDVEl'
            b'MqkAfp6qS/8xcGBTEqn08dDQIgPl8KofpC9GXMGbK\nV9FGP+c1bpA3vMOfnpsE'
            b'WCju2qDoTSKJTm3VMZj88mqH7nOpbk7JI/Yz0EmtNXOM\n6QIDAQAB\n-----EN'
            b'D PUBLIC KEY-----')

    def save_password(self, password):
        """Store sudo password"""
        # TODO: REMOTE API RAPI: Remove this API on the next RAPI bump
        # if the slave is running it means we don't need password
        # so we can consider call to this function as passing
        return True

    def finish_job(self, result=None):
        # assert the thread completed
        self.session_change_lock.acquire(blocking=False)
        self.session_change_lock.release()
        if self._sa.get_job(self._currently_running_job).plugin in [
                'manual', 'user-interact-verify'] and not result:
            # for manually verified jobs we don't set the outcome here
            # it is already determined
            return
        if not result:
            result = self._be.wait().get_result()
        self._sa.use_job_result(self._currently_running_job, result)
        if self._state != Bootstrapping:
            if not self._sa.get_dynamic_todo_list():
                if (
                    self._launcher.auto_retry and
                    self.get_rerun_candidates('auto')
                ):
                    self._state = TestsSelected
                else:
                    self._state = Idle
            else:
                self._state = TestsSelected
        return result

    def get_rerun_candidates(self, session_type='manual'):
        return self._sa.get_rerun_candidates(session_type)

    def prepare_rerun_candidates(self, rerun_candidates):
        candidates = self._sa.prepare_rerun_candidates(rerun_candidates)
        self._state = TestsSelected
        return candidates

    def get_job_result(self, job_id):
        return self._sa.get_job_state(job_id).result

    def get_jobs_repr(self, job_ids, offset=0):
        """
        Translate jobs into a {'field': 'val'} representations.

        :param job_ids:
            list of job ids to get and translate
        :param offset:
            apply an offset to the job number if for instance the job list
            is being requested part way through a session
        :returns:
            list of dicts representing jobs
        """
        test_info_list = tuple()
        for job_no, job_id in enumerate(job_ids, start=offset + 1):
            job = self._sa.get_job(job_id)
            cat_id = self._sa.get_job_state(job.id).effective_category_id
            duration_txt = _('No estimated duration provided for this job')
            if job.estimated_duration is not None:
                duration_txt = '{} {}'.format(job.estimated_duration, _(
                    'seconds'))
            # the next dict is only to get test_info generating code tidier
            automated_desc = {
                True: _('this job is fully automated'),
                False: _('this job requires some manual interaction')
            }
            test_info = {
                "id": job.id,
                "partial_id": job.partial_id,
                "name": job.tr_summary(),
                "category_id": cat_id,
                "category_name": self._sa.get_category(cat_id).tr_name(),
                "automated": automated_desc[job.automated],
                "duration": duration_txt,
                "description": (job.tr_description() or
                                _('No description provided for this job')),
                "outcome": self._sa.get_job_state(job.id).result.outcome,
                "user": job.user,
                "command": job.command,
                "num": job_no,
                "plugin": job.plugin,
            }
            test_info_list = test_info_list + ((test_info, ))
        return json.dumps(test_info_list)

    def resume_by_id(self, session_id=None):
        _logger.info("resume_by_id: %r", session_id)
        self._launcher = load_configs()
        resume_candidates = list(self._sa.get_resumable_sessions())
        if not session_id:
            if not resume_candidates:
                print('No session to resume')
                return
            session_id = resume_candidates[0].id
        if session_id not in [s.id for s in resume_candidates]:
            print("Requested session not found")
            return
        _logger.warning("Resuming session: %r", session_id)
        runner_kwargs = {
            'normal_user_provider': lambda: self._normal_user,
            'stdin': self._pipe_to_subproc,
            'extra_env': self.prepare_extra_env,
        }
        meta = self._sa.resume_session(session_id, runner_kwargs=runner_kwargs)
        app_blob = json.loads(meta.app_blob.decode("UTF-8"))
        launcher = app_blob['launcher']
        self._launcher.read_string(launcher, False)
        self._sa.use_alternate_configuration(self._launcher)

        self._normal_user = app_blob.get(
            'effective_normal_user', self._launcher.normal_user)
        _logger.info(
            "normal_user after loading metadata: %r", self._normal_user)
        test_plan_id = app_blob['testplan_id']
        self._sa.select_test_plan(test_plan_id)
        self._sa.bootstrap()
        self._last_job = meta.running_job_name

        result_dict = {
            'outcome': IJobResult.OUTCOME_PASS,
            'comments': _("Automatically passed after resuming execution"),
        }
        session_share = WellKnownDirsHelper.session_share(
            self._sa._manager.storage.id)
        result_path = os.path.join(session_share, '__result')
        if os.path.exists(result_path):
            try:
                with open(result_path, 'rt') as f:
                    result_dict = json.load(f)
                    # the only really important field in the result is
                    # 'outcome' so let's make sure it doesn't contain
                    # anything stupid
                    if result_dict.get('outcome') not in [
                            'pass', 'fail', 'skip']:
                        result_dict['outcome'] = IJobResult.OUTCOME_PASS
            except json.JSONDecodeError:
                pass
        result = MemoryJobResult(result_dict)
        if self._last_job:
            try:
                self._sa.use_job_result(self._last_job, result, True)
            except KeyError:
                raise SystemExit(self._last_job)

        # some jobs have already been run, so we need to update the attempts
        # count for future auto-rerunning
        if self._launcher.auto_retry:
            for job_id in [
                    job.id for job in self.get_rerun_candidates('auto')]:
                job_state = self._sa.get_job_state(job_id)
                job_state.attempts = self._launcher.max_attempts - len(
                    job_state.result_history)

        self._state = TestsSelected

    def finalize_session(self):
        self._sa.finalize_session()
        self._reset_sa()

    def transmit_input(self, text):
        if not text:
            self._pipe_from_master.close()
            return
        self._pipe_from_master.write(text)
        self._pipe_from_master.flush()

    def send_signal(self, signal):
        if not self._currently_running_job:
            return
        target_user = self._sa.get_job(self._currently_running_job).user
        self._sa.send_signal(signal, target_user)

    @property
    def manager(self):
        return self._sa._manager

    @property
    def passwordless_sudo(self):
        # TODO: REMOTE API RAPI: Remove this API on the next RAPI bump
        # if the slave is still running it means it's very passwordless
        return True

    @property
    def sideloaded_providers(self):
        return self._sa.sideloaded_providers

    def exposed_cache_report(self, exporter_id, options):
        exporter = self._sa._manager.create_exporter(exporter_id, options)
        exported_stream = SpooledTemporaryFile(max_size=102400, mode='w+b')
        exporter.dump_from_session_manager(self._sa._manager, exported_stream)
        exported_stream.flush()
        return exported_stream

def _guess_normal_user():
    _logger.warning("normal_user not supplied via config(s).")
    for entry in pwd.getpwall():
        if entry.pw_name == 'ubuntu':
            _logger.warning("Using `ubuntu` user")
            return 'ubuntu'
    with suppress(KeyError):
        user = pwd.getpwuid(1000).pw_name
        _logger.warning("Using `%s` user", user)
        return user
    with suppress(KeyError):
        user = pwd.getpwuid(1001).pw_name
        _logger.warning("Using `%s` user", user)
        return user
    raise RuntimeError(
        ("normal_user not supplied via config(s). "
            "User `ubuntu` and username for uid 1000 or 1001 were not found"))



# This file is part of Checkbox.
#
# Copyright 2012-2014 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.
"""
plainbox.impl.commands.test_dev
===============================

Test definitions for plainbox.impl.dev module
"""
import argparse
from inspect import cleandoc
from unittest import TestCase

from plainbox.impl.commands.dev import DevCommand
from plainbox.testing_utils.argparse_compat import optionals_section
from plainbox.testing_utils.io import TestIO
from plainbox.vendor import mock


class TestDevCommand(TestCase):

    def setUp(self):
        self.parser = argparse.ArgumentParser(prog='test')
        self.subparsers = self.parser.add_subparsers()
        self.provider_loader = lambda: [mock.Mock()]
        self.config_loader = lambda: mock.Mock()
        self.ns = mock.Mock()

    def test_init(self):
        dev_cmd = DevCommand(self.provider_loader, self.config_loader)
        self.assertIs(dev_cmd.provider_loader, self.provider_loader)
        self.assertIs(dev_cmd.config_loader, self.config_loader)

    def test_register_parser(self):
        DevCommand(self.provider_loader, self.config_loader).register_parser(
            self.subparsers)
        with TestIO() as io:
            self.parser.print_help()
        self.assertIn("development commands", io.stdout)
        with TestIO() as io:
            with self.assertRaises(SystemExit):
                self.parser.parse_args(['dev', '--help'])
        self.maxDiff = None
        self.assertEqual(
            io.stdout, cleandoc(
                """
                usage: plainbox dev <subcommand> ...

                positional arguments:
                  {{parse}}
                    parse     parse stdin with the specified parser

                {}:
                  -h, --help  show this help message and exit
                """.format(optionals_section))
            + "\n")



# This file is part of Checkbox.
#
# Copyright 2012, 2013 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
plainbox.impl.test_ctrl
=======================

Test definitions for plainbox.impl.ctrl module
"""

from subprocess import CalledProcessError
from unittest import TestCase
import os
import shutil

from plainbox.abc import IJobResult
from plainbox.abc import IProvider1
from plainbox.abc import IProviderBackend1
from plainbox.impl.ctrl import CheckBoxSessionStateController
from plainbox.impl.ctrl import SymLinkNest
from plainbox.impl.ctrl import gen_rfc822_records_from_io_log
from plainbox.impl.job import JobDefinition
from plainbox.impl.resource import Resource
from plainbox.impl.resource import ResourceExpression
from plainbox.impl.result import MemoryJobResult
from plainbox.impl.secure.origin import JobOutputTextSource
from plainbox.impl.secure.origin import Origin
from plainbox.impl.secure.providers.v1 import Provider1
from plainbox.impl.secure.rfc822 import RFC822Record
from plainbox.impl.secure.rfc822 import RFC822SyntaxError
from plainbox.impl.session import InhibitionCause
from plainbox.impl.session import JobReadinessInhibitor
from plainbox.impl.session import JobState
from plainbox.impl.session import SessionState
from plainbox.impl.testing_utils import make_job
from plainbox.impl.unit.template import TemplateUnit
from plainbox.vendor import extcmd
from plainbox.vendor import mock


class CheckBoxSessionStateControllerTests(TestCase):

    def setUp(self):
        self.ctrl = CheckBoxSessionStateController()

    def test_get_dependency_set(self):
        # Job with no dependencies
        job_a = JobDefinition({})
        self.assertEqual(
            self.ctrl.get_dependency_set(job_a), set())
        # Job with direct dependencies
        job_b = JobDefinition({
            'depends': 'j1, j2'
        })
        self.assertEqual(
            self.ctrl.get_dependency_set(job_b),
            {('direct', 'j1'), ('direct', 'j2')})
        # Job with resouce dependencies
        job_c = JobDefinition({
            'requires': 'j3.attr == 1'
        })
        self.assertEqual(
            self.ctrl.get_dependency_set(job_c),
            {('resource', 'j3')})
        # Job with ordering dependencies
        job_d = JobDefinition({
            'after': 'j1, j2'
        })
        self.assertEqual(
            self.ctrl.get_dependency_set(job_d),
            {('ordering', 'j1'), ('ordering', 'j2')})
        # Job with both direct and resource dependencies
        job_e = JobDefinition({
            'depends': 'j4',
            'requires': 'j5.attr == 1'
        })
        self.assertEqual(
            self.ctrl.get_dependency_set(job_e),
            {('direct', 'j4'), ('resource', 'j5')})
        # Job with both direct and resource dependencies
        # on the same job (j6)
        job_f = JobDefinition({
            'depends': 'j6',
            'requires': 'j6.attr == 1'
        })
        self.assertEqual(
            self.ctrl.get_dependency_set(job_f),
            {('direct', 'j6'), ('resource', 'j6')})

    def test_get_inhibitor_list_PENDING_RESOURCE(self):
        # verify that jobs that require a resource that hasn't been
        # invoked yet produce the PENDING_RESOURCE inhibitor
        j1 = JobDefinition({
            'id': 'j1',
            'requires': 'j2.attr == "ok"'
        })
        j2 = JobDefinition({
            'id': 'j2'
        })
        session_state = mock.MagicMock(spec=SessionState)
        session_state.job_state_map['j2'].job = j2
        session_state.resource_map = {}
        self.assertEqual(
            self.ctrl.get_inhibitor_list(session_state, j1),
            [JobReadinessInhibitor(
                InhibitionCause.PENDING_RESOURCE,
                j2, ResourceExpression('j2.attr == "ok"'))])

    def test_get_inhibitor_list_FAILED_RESOURCE(self):
        # verify that jobs that require a resource that has been
        # invoked and produced resources but the expression dones't
        # evaluate to True produce the FAILED_RESOURCE inhibitor
        j1 = JobDefinition({
            'id': 'j1',
            'requires': 'j2.attr == "ok"'
        })
        j2 = JobDefinition({
            'id': 'j2'
        })
        session_state = mock.MagicMock(spec=SessionState)
        session_state.job_state_map['j2'].job = j2
        session_state.resource_map = {
            'j2': [Resource({'attr': 'not-ok'})]
        }
        self.assertEqual(
            self.ctrl.get_inhibitor_list(session_state, j1),
            [JobReadinessInhibitor(
                InhibitionCause.FAILED_RESOURCE,
                j2, ResourceExpression('j2.attr == "ok"'))])

    def test_get_inhibitor_list_good_resource(self):
        # verify that jobs that require a resource that has been invoked and
        # produced resources for which the expression evaluates to True don't
        # have any inhibitors
        j1 = JobDefinition({
            'id': 'j1',
            'requires': 'j2.attr == "ok"'
        })
        j2 = JobDefinition({
            'id': 'j2'
        })
        session_state = mock.MagicMock(spec=SessionState)
        session_state.resource_map = {
            'j2': [Resource({'attr': 'ok'})]
        }
        session_state.job_state_map['j2'].job = j2
        self.assertEqual(
            self.ctrl.get_inhibitor_list(session_state, j1), [])

    def test_get_inhibitor_list_PENDING_DEP(self):
        # verify that jobs that depend on another job or wait (via after) for
        # another  that hasn't been invoked yet produce the PENDING_DEP
        # inhibitor
        j1 = JobDefinition({
            'id': 'j1',
            'depends': 'j2',
            'after': 'j3',
        })
        j2 = JobDefinition({
            'id': 'j2'
        })
        j3 = JobDefinition({
            'id': 'j3'
        })
        session_state = mock.MagicMock(spec=SessionState)
        session_state.job_state_map = {
            'j1': mock.Mock(spec_set=JobState),
            'j2': mock.Mock(spec_set=JobState),
            'j3': mock.Mock(spec_set=JobState),
        }
        jsm_j2 = session_state.job_state_map['j2']
        jsm_j2.job = j2
        jsm_j2.result.outcome = IJobResult.OUTCOME_NONE
        jsm_j3 = session_state.job_state_map['j3']
        jsm_j3.job = j3
        jsm_j3.result.outcome = IJobResult.OUTCOME_NONE
        self.assertEqual(self.ctrl.get_inhibitor_list(session_state, j1), [
            JobReadinessInhibitor(InhibitionCause.PENDING_DEP, j2, None),
            JobReadinessInhibitor(InhibitionCause.PENDING_DEP, j3, None),
        ])

    def test_get_inhibitor_list_FAILED_DEP(self):
        # verify that jobs that depend on another job that ran but
        # didn't result in OUTCOME_PASS produce the FAILED_DEP
        # inhibitor.
        j1 = JobDefinition({
            'id': 'j1',
            'depends': 'j2',
            'after': 'j3',
        })
        j2 = JobDefinition({
            'id': 'j2'
        })
        j3 = JobDefinition({
            'id': 'j3'
        })
        session_state = mock.MagicMock(spec=SessionState)
        session_state.job_state_map = {
            'j1': mock.Mock(spec_set=JobState),
            'j2': mock.Mock(spec_set=JobState),
            'j3': mock.Mock(spec_set=JobState),
        }
        jsm_j2 = session_state.job_state_map['j2']
        jsm_j2.job = j2
        jsm_j2.result.outcome = IJobResult.OUTCOME_FAIL
        jsm_j3 = session_state.job_state_map['j3']
        jsm_j3.job = j3
        jsm_j3.result.outcome = IJobResult.OUTCOME_FAIL
        self.assertEqual(
            self.ctrl.get_inhibitor_list(session_state, j1),
            [JobReadinessInhibitor(
                InhibitionCause.FAILED_DEP, j2, None)])

    def test_get_inhibitor_list_good_dep(self):
        # verify that jobs that depend on another job that ran and has outcome
        # equal to OUTCOME_PASS don't have any inhibitors
        j1 = JobDefinition({
            'id': 'j1',
            'depends': 'j2',
            'after': 'j3'
        })
        j2 = JobDefinition({
            'id': 'j2'
        })
        j3 = JobDefinition({
            'id': 'j3'
        })
        session_state = mock.MagicMock(spec=SessionState)
        session_state.job_state_map = {
            'j1': mock.Mock(spec_set=JobState),
            'j2': mock.Mock(spec_set=JobState),
            'j3': mock.Mock(spec_set=JobState),
        }
        jsm_j2 = session_state.job_state_map['j2']
        jsm_j2.job = j2
        jsm_j2.result.outcome = IJobResult.OUTCOME_PASS
        jsm_j3 = session_state.job_state_map['j3']
        jsm_j3.job = j3
        jsm_j3.result.outcome = IJobResult.OUTCOME_PASS
        self.assertEqual(
            self.ctrl.get_inhibitor_list(session_state, j1), [])

    def test_observe_result__normal(self):
        job = mock.Mock(spec=JobDefinition)
        result = mock.Mock(spec=IJobResult)
        session_state = mock.MagicMock(spec=SessionState)
        self.ctrl.observe_result(session_state, job, result)
        # Ensure that result got stored
        self.assertIs(
            session_state.job_state_map[job.id].result, result)
        # Ensure that signals got fired
        session_state.on_job_state_map_changed.assert_called_once_with()
        session_state.on_job_result_changed.assert_called_once_with(
            job, result)

    def test_observe_result__OUTCOME_NONE(self):
        job = mock.Mock(spec=JobDefinition, plugin='resource')
        result = mock.Mock(spec=IJobResult, outcome=IJobResult.OUTCOME_NONE)
        session_state = mock.MagicMock(spec=SessionState)
        self.ctrl.observe_result(session_state, job, result)
        # Ensure that result got stored
        self.assertIs(
            session_state.job_state_map[job.id].result, result)
        # Ensure that signals got fired
        session_state.on_job_state_map_changed.assert_called_once_with()
        session_state.on_job_result_changed.assert_called_once_with(
            job, result)
        # Ensure that a resource was *not* defined
        self.assertEqual(session_state.set_resource_list.call_count, 0)

    def test_observe_result__resource(self):
        job = mock.Mock(spec=JobDefinition, plugin='resource')
        result = mock.Mock(spec=IJobResult, outcome=IJobResult.OUTCOME_PASS)
        result.get_io_log.return_value = [
            (0, 'stdout', b'attr: value1\n'),
            (0, 'stdout', b'\n'),
            (0, 'stdout', b'attr: value2\n')]
        session_state = mock.MagicMock(spec=SessionState)
        self.ctrl.observe_result(session_state, job, result)
        # Ensure that result got stored
        self.assertIs(
            session_state.job_state_map[job.id].result, result)
        # Ensure that signals got fired
        session_state.on_job_state_map_changed.assert_called_once_with()
        session_state.on_job_result_changed.assert_called_once_with(
            job, result)
        # Ensure that new resource was defined
        session_state.set_resource_list.assert_called_once_with(
            job.id, [
                Resource({'attr': 'value1'}), Resource({'attr': 'value2'})])

    @mock.patch('plainbox.impl.ctrl.logger')
    def test_observe_result__broken_resource(self, mock_logger):
        job = mock.Mock(spec=JobDefinition, plugin='resource')
        result = mock.Mock(spec=IJobResult, outcome=IJobResult.OUTCOME_PASS)
        result.get_io_log.return_value = [(0, 'stdout', b'barf\n')]
        session_state = mock.MagicMock(spec=SessionState)
        self.ctrl.observe_result(session_state, job, result)
        # Ensure that result got stored
        self.assertIs(
            session_state.job_state_map[job.id].result, result)
        # Ensure that signals got fired
        session_state.on_job_state_map_changed.assert_called_once_with()
        session_state.on_job_result_changed.assert_called_once_with(
            job, result)
        # Ensure that a warning was logged
        mock_logger.warning.assert_called_once_with(
            "local script %s returned invalid RFC822 data: %s",
            job.id, RFC822SyntaxError(
                None, 1, "Unexpected non-empty line: 'barf\\n'"))

    @mock.patch('plainbox.impl.ctrl.logger')
    def test_observe_result__missing_resource_key(self, mock_logger):
        job = make_job("R", plugin="resource")
        template = TemplateUnit({
            'template-resource': job.id,
            'id': 'foo-{missing}',
            'plugin': 'shell'})
        result = mock.Mock(spec=IJobResult, outcome=IJobResult.OUTCOME_PASS)
        result.get_io_log.return_value = [
            (0, 'stdout', b'attr: value1\n'),
            (0, 'stdout', b'\n'),
            (0, 'stdout', b'attr: value2\n')]
        session_state = SessionState([template, job])
        self.ctrl.observe_result(session_state, job, result)
        # Ensure that a warning was logged
        mock_logger.debug.assert_called_with(
            "Ignoring %s with missing template parameter %s",
            "foo-{missing}", "missing")


class FunctionTests(TestCase):
    """
    unit tests for gen_rfc822_records_from_io_log() and other functions.
    """

    def test_parse_typical(self):
        """
        verify typical operation without any parsing errors
        """
        # Setup a mock job and result, give some io log to the result
        job = mock.Mock(spec=JobDefinition)
        result = mock.Mock(spec=IJobResult)
        result.get_io_log.return_value = [
            (0, 'stdout', b'attr: value1\n'),
            (0, 'stdout', b'\n'),
            (0, 'stdout', b'attr: value2\n')]
        # Parse the IO log records
        records = list(gen_rfc822_records_from_io_log(job, result))
        # Ensure that we saw both records
        self.assertEqual(records, [
            RFC822Record(
                {'attr': 'value1'}, Origin(JobOutputTextSource(job), 1, 1)),
            RFC822Record(
                {'attr': 'value2'}, Origin(JobOutputTextSource(job), 3, 3)),
        ])

    @mock.patch('plainbox.impl.ctrl.logger')
    def test_parse_error(self, mock_logger):
        # Setup a mock job and result, give some io log to the result
        job = mock.Mock(spec=JobDefinition)
        result = mock.Mock(spec=IJobResult)
        result.get_io_log.return_value = [
            (0, 'stdout', b'attr: value1\n'),
            (0, 'stdout', b'\n'),
            (0, 'stdout', b'error\n'),
            (0, 'stdout', b'\n'),
            (0, 'stdout', b'attr: value2\n')]
        # Parse the IO log records
        records = list(gen_rfc822_records_from_io_log(job, result))
        # Ensure that only the first record was generated
        self.assertEqual(records, [
            RFC822Record(
                {'attr': 'value1'}, Origin(JobOutputTextSource(job), 1, 1)),
        ])
        # Ensure that a warning was logged
        mock_logger.warning.assert_called_once_with(
            "local script %s returned invalid RFC822 data: %s",
            job.id, RFC822SyntaxError(
                None, 3, "Unexpected non-empty line: 'error\\n'"))


class SymLinkNestTests(TestCase):
    """
    Tests for SymLinkNest class
    """

    NEST_DIR = "nest"

    def setUp(self):
        self.nest = SymLinkNest(self.NEST_DIR)

    def test_init(self):
        """
        verify that SymLinkNest.__init__() stores its argument
        """
        self.assertEqual(self.nest._dirname, self.NEST_DIR)

    def test_add_provider(self):
        """
        verify that add_provider() adds each executable
        """
        provider = mock.Mock(name='provider', spec=Provider1)
        provider.executable_list = ['exec1', 'exec2']
        with mock.patch.object(self.nest, 'add_executable'):
            self.nest.add_provider(provider)
            self.nest.add_executable.assert_has_calls([
                (('exec1',), {}),
                (('exec2',), {})])

    @mock.patch('os.symlink')
    def test_add_executable(self, mock_symlink):
        self.nest.add_executable('/usr/lib/foo/exec')
        mock_symlink.assert_called_with(
            '/usr/lib/foo/exec', 'nest/exec')
